from marshmallow import (
    fields,
    RAISE,
)

from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_lib_root.enum_desc.CompScope import CompScope
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc

client_version_ = "client_version"
client_conf_target_ = "client_conf_target"
server_action_ = "server_action"
command_line_ = "command_line"
cursor_cpos_ = "cursor_cpos"
comp_scope_ = "comp_scope"
client_uid_ = "client_uid"
client_pid_ = "client_pid"
is_debug_enabled_ = "is_debug_enabled"
input_data_ = "input_data"

_sample_command_line = "some_command goto service "

_call_context_example = {
    client_version_: "",
    client_conf_target_: "",
    server_action_: ServerAction.DescribeLineArgs.name,
    command_line_: _sample_command_line,
    cursor_cpos_: len(_sample_command_line),
    comp_scope_: CompScope.ScopeInitial.name,
    client_uid_: "",
    client_pid_: 0,
    is_debug_enabled_: False,
    input_data_: None,
}


class CallContextSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        ordered = True

    model_class = CallContext

    client_version = fields.String(
        required = False,
        metadata = {
            "description": "client version of `argrelay` package (empty = not set)",
            "example": _call_context_example[client_version_],
        },
        load_default = "",
    )
    client_conf_target = fields.String(
        required = False,
        metadata = {
            "description": "target path of `@/conf` symlink (empty = not set)",
            "example": _call_context_example[client_conf_target_],
        },
        load_default = "",
    )
    server_action = fields.Enum(
        ServerAction,
        required = True,
        metadata = {
            "description": (
                "Action to perform - see " + ServerAction.__name__ + " enum"
            ),
            "example": _call_context_example[server_action_],
        },
    )
    command_line = fields.String(
        required = True,
        metadata = {
            "example": _call_context_example[command_line_],
        },
    )
    cursor_cpos = fields.Integer(
        required = True,
        metadata = {
            "description": "Cursor position within command line (0 = before the first char)",
            "example": _call_context_example[cursor_cpos_],
        },
    )
    comp_scope = fields.Enum(
        CompScope,
        required = True,
        metadata = {
            "description": "Name for a completion scope - see " + CompScope.__name__ + " enum",
            "example": _call_context_example[comp_scope_],
        },
    )
    client_uid = fields.String(
        required = False,
        metadata = {
            "description": "UID (user name) of the client user according to OS (empty = not set)",
            "example": _call_context_example[client_uid_],
        },
        load_default = "",
    )
    client_pid = fields.Integer(
        required = False,
        metadata = {
            "description": "PID of the client process local to the client host (0 = not set)",
            "example": _call_context_example[client_pid_],
        },
        load_default = 0,
    )
    is_debug_enabled = fields.Boolean(
        required = True,
        metadata = {
            "description": "Enable extra debug output",
            "example": _call_context_example[is_debug_enabled_],
        },
    )
    input_data = fields.String(
        required = False,
        load_default = None,
        metadata = {
            "description": "Enable extra debug output",
            "example": _call_context_example[input_data_],
        },
    )


call_context_desc = TypeDesc(
    dict_schema = CallContextSchema(),
    ref_name = CallContextSchema.__name__,
    dict_example = _call_context_example,
    default_file_path = "",
)



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry


@dataclass
class CheckEnvPluginConfig:
    """
    See also `CheckEnvPluginConfigSchema`.
    """

    check_env_plugin_instances: dict[str, PluginEntry] = field()
    """
    Same as `server_plugin_instances` but for FS_36_17_84_44 `check_env` only.
    """

    check_env_plugin_instance_id_activate_list: list[str] = field()
    """
    Same as `plugin_instance_id_activate_list` but for FS_36_17_84_44 `check_env` only.
    """



from __future__ import annotations

from marshmallow import (
    fields,
    RAISE,
    Schema,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_server_plugin_demo.demo_git.GitRepoEntryConfigSchema import git_repo_entry_config_desc
from argrelay_lib_server_plugin_demo.demo_git.GitRepoEnvelopeClass import GitRepoEnvelopeClass

class_name_repo_ = "class_name_repo"
class_name_tag_ = "class_name_tag"
class_name_commit_ = "class_name_commit"

load_git_repos_default_ = "load_git_repos_default"
load_git_tags_default_ = "load_git_tags_default"
load_git_commits_default_ = "load_git_commits_default"

repo_entries_ = "repo_entries"


class GitRepoLoaderConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    class_name_repo = fields.String(
        required = False,
        load_default = GitRepoEnvelopeClass.class_git_repo.name,
    )

    class_name_tag = fields.String(
        required = False,
        load_default = GitRepoEnvelopeClass.class_git_tag.name,
    )

    class_name_commit = fields.String(
        required = False,
        load_default = GitRepoEnvelopeClass.class_git_commit.name,
    )

    load_git_repos_default = fields.Boolean(
        required = False,
        load_default = True,
    )

    load_git_tags_default = fields.Boolean(
        required = False,
        load_default = False,
    )

    load_git_commits_default = fields.Boolean(
        required = False,
        load_default = False,
    )

    # Maps `repo_base_path` to list of repo entries.
    # *   If `repo_base_path` starts with `/` (absolute), it is used verbatim.
    # *   If `repo_base_path` does not start with `/` (relative),
    #     it is converted to absolute assuming it is relative to `argrelay_dir` (`@/`).
    repo_entries = fields.Dict(
        keys = fields.String(),
        values = fields.List(
            fields.Nested(git_repo_entry_config_desc.dict_schema),
            required = True,
        ),
        required = True,
    )


git_repo_loader_config_desc = TypeDesc(
    dict_schema = GitRepoLoaderConfigSchema(),
    ref_name = GitRepoLoaderConfigSchema.__name__,
    dict_example = {
        class_name_repo_: GitRepoEnvelopeClass.class_git_repo.name,
        class_name_tag_: GitRepoEnvelopeClass.class_git_tag.name,
        class_name_commit_: GitRepoEnvelopeClass.class_git_commit.name,
        load_git_repos_default_: True,
        load_git_tags_default_: False,
        load_git_commits_default_: False,
        repo_entries_: {
            "~/repos": [
                git_repo_entry_config_desc.dict_example,
            ]
        },
    },
    default_file_path = "",
)



from __future__ import annotations

from copy import deepcopy
from dataclasses import (
    dataclass,
    field,
)
from typing import Union

from argrelay_app_server.relay_server import HelpHintCache
from argrelay_app_server.relay_server.QueryEngine import (
    populate_query_dict,
    QueryEngine,
)
from argrelay_app_server.relay_server.QueryResult import QueryResult
from argrelay_app_server.runtime_context.AbstractArg import (
    ArgCommand,
    ArgCommandIncomplete,
    ArgCommandValueDictated,
    ArgCommandValueOffered,
)
from argrelay_app_server.runtime_context.DataArg import (
    ArgCommandDataIncomplete,
    ArgCommandDataValueDictated,
    ArgCommandDataValueOffered,
)
from argrelay_app_server.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay_app_server.runtime_context.ParsedContext import ParsedContext
from argrelay_app_server.runtime_context.SearchControl import SearchControl
from argrelay_app_server.runtime_context.token_bucket_utils import token_buckets_to_token_ipos_list
from argrelay_lib_root.enum_desc.InterpStep import InterpStep
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.enum_desc.SpecialChar import SpecialChar
from argrelay_lib_root.enum_desc.TermColor import TermColor
from argrelay_lib_root.enum_desc.ValueSource import ValueSource
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_lib_root.misc_helper_common.ElapsedTime import ElapsedTime

function_container_ipos_ = 0


@dataclass
class InterpContext:
    """
    Mutable state for the process of command line interpretation.
    """

    parsed_ctx: ParsedContext = field()

    # TODO: Move all dynamic and non-serializable objects into `InterpRuntime` (or something like that).
    interp_factories: dict[str, "AbstractInterpFactory"] = field()
    """
    Reference to `ServerConfig.action_delegators`.
    """

    action_delegators: dict[str, "DelegatorAbstract"] = field()
    """
    Reference to `ServerConfig.action_delegators`.
    """

    query_engine: QueryEngine = field()

    help_hint_cache: HelpHintCache = field()

    excluded_tokens: list[int] = field(init = False, default_factory = lambda: [])
    """
    Tokens excluded by ways other than consumption into `consumed_token_buckets`.
    """

    last_token_bucket_used: Union[int, None] = None
    """
    FS_97_64_39_94: ipos of the `token_bucket` which was used by one of the `EnvelopeContainer`-s
    """

    included_token_buckets: list[list[int]] = field(init = False, default_factory = lambda: [])
    """
    FS_97_64_39_94: `token_bucket`-s
    
    If `included_token_buckets` are combined,
    the result will contain both `remaining_token_buckets` and `consumed_token_buckets`.
    Field `included_token_buckets` is a maximum set for `token_bucket`-s -
    it is similar to the maximum set `ParsedContext.all_tokens`,
    but it cannot contain all tokens - it must exclude at least `SpecialChar.TokenBucketDelimiter` to start with.
    """

    remaining_token_buckets: list[list[int]] = field(init = False)
    """
    Same as `included_token_buckets` but for remaining tokens only.
    """

    consumed_token_buckets: list[list[int]] = field(init = False, default_factory = lambda: [])
    """
    Same as `included_token_buckets` but for consumed tokens only.
    """

    token_ipos_to_token_bucket_map: dict[int, int] = field(init = False, default_factory = lambda: {})
    """
    Index reversed to `included_token_buckets` -
    for each `token_ipos` it gives the index of the `token_bucket` it is in.
    It does not map `token_ipos`-es from `excluded_tokens`.
    """

    remaining_dictated_args_per_bucket: list[list[ArgCommandValueDictated]] = field(
        init = False,
        default_factory = lambda: [],
    )
    """
    Remaining `dictated_arg`-s for each bucket.
    """

    remaining_offered_args_per_bucket: list[list[ArgCommandValueOffered]] = field(
        init = False,
        default_factory = lambda: [],
    )
    """
    Remaining `offered_arg`-s for each bucket.
    """

    remaining_incomplete_args_per_bucket: list[list[ArgCommandIncomplete]] = field(
        init = False,
        default_factory = lambda: [],
    )
    """
    Remaining FS_08_58_30_24 `incomplete_arg`-s for each bucket.
    """

    tangent_arg: Union[ArgCommand, None] = field(init = False, default = None)
    """
    Similar to `tangent_token` but affects entire `command_arg` this `tangent_token` is part of.
    """

    token_ipos_to_arg_map: dict[int, ArgCommand] = field(init = False, default_factory = lambda: {})
    """
    Maps `token_ipos` into one of the args in
    either `remaining_offered_args_per_bucket` or `remaining_offered_args_per_bucket` (or `None`).
    """

    envelope_containers: list[EnvelopeContainer] = field(init = False, default_factory = lambda: [])
    """
    Each `envelope_container` wraps `data_envelope`-s matching query and some associated data.
    """

    curr_container: EnvelopeContainer = field(init = False, default = None)
    """
    One of the `envelope_containers` currently being searched.
    """

    curr_container_ipos: int = field(init = False, default = -1)
    """
    It is an `ipos` into `envelope_containers` to select `curr_container` (what is currently being searched for).

    Check `EnvelopeContainer.found_count` to see whether this container is singled out or not.

    All `envelope_containers` after `curr_container_ipos` will not be pointing to any `data_envelope`-s
    (search procedure has not reached them yet).
    """

    prev_interp: "AbstractInterp" = field(init = False, default = None)

    curr_interp: "AbstractInterp" = field(init = False, default = None)
    """
    Current interpreter during command line interpretation.
    """

    comp_suggestions: list[str] = field(init = False, default_factory = lambda: [])
    """
    Sorted list of completion suggestions.
    """

    interp_tree_abs_path: tuple[str, ...] = field(init = False, default = tuple([]))
    """
    Provides curr path within FS_01_89_09_24 interp tree.
    Takes part in implementation of FS_01_89_09_24 interp tree.
    """

    def __post_init__(self):
        self._init_token_buckets()
        self._build_arg_collections()

    def _init_token_buckets(self):
        self.included_token_buckets.append([])
        curr_bucket_index = len(self.included_token_buckets) - 1
        curr_token_bucket = self.included_token_buckets[curr_bucket_index]
        for token_ipos in range(0, len(self.parsed_ctx.all_tokens)):

            # Split and populate FS_97_64_39_94 `token_bucket`-s:
            if self.parsed_ctx.all_tokens[token_ipos] == SpecialChar.TokenBucketDelimiter.value:
                self.included_token_buckets.append([])
                curr_bucket_index = len(self.included_token_buckets) - 1
                curr_token_bucket = self.included_token_buckets[curr_bucket_index]
                # Exclude `SpecialChar.TokenBucketDelimiter` unconditionally:
                self.excluded_tokens.append(token_ipos)
            else:
                curr_token_bucket.append(token_ipos)
                self.token_ipos_to_token_bucket_map[token_ipos] = curr_bucket_index

        # Init remaining and consumed:
        self.remaining_token_buckets = deepcopy(self.included_token_buckets)
        for i in range(len(self.included_token_buckets)):
            self.consumed_token_buckets.append([])

    def _build_arg_collections(
        self,
    ):
        """
        Interpret `command_token`-s as `command_arg`-s within `token_bucket`-s
        """

        all_tokens = self.parsed_ctx.all_tokens

        for token_bucket in self.remaining_token_buckets:

            offered_args: list[ArgCommandValueOffered] = []
            dictated_args: list[ArgCommandValueDictated] = []
            incomplete_args: list[ArgCommandIncomplete] = []
            self.remaining_offered_args_per_bucket.append(offered_args)
            self.remaining_dictated_args_per_bucket.append(dictated_args)
            self.remaining_incomplete_args_per_bucket.append(incomplete_args)

            arg_name_ipos: Union[int, None] = None
            arg_name_value: Union[str, None] = None
            for token_ipos in token_bucket:

                token_value = all_tokens[token_ipos]

                if arg_name_value is None:
                    if token_value.startswith(SpecialChar.ArgNamePrefix.value):
                        # Skip `SpecialChar.ArgNamePrefix`:
                        arg_name_value = token_value[1:]
                        arg_name_ipos = token_ipos
                        continue
                    else:
                        command_arg = ArgCommandDataValueOffered(
                            token_ipos_list = [token_ipos],
                            arg_value = token_value,
                        )
                        offered_args.append(command_arg)
                        self.token_ipos_to_arg_map[token_ipos] = command_arg
                else:
                    command_arg = ArgCommandDataValueDictated(
                        token_ipos_list = [arg_name_ipos, token_ipos],
                        arg_name = arg_name_value,
                        arg_value = token_value,
                    )
                    dictated_args.append(command_arg)
                    self.token_ipos_to_arg_map[arg_name_ipos] = command_arg
                    self.token_ipos_to_arg_map[token_ipos] = command_arg
                    arg_name_ipos = None
                    arg_name_value = None

            # If `dictated_arg` is incomplete (has `arg_name`, but missing `arg_value`),
            # it is registered as `incomplete_arg`.
            if arg_name_value is not None:
                command_arg = ArgCommandDataIncomplete(
                    token_ipos_list = [arg_name_ipos],
                    arg_name = arg_name_value,
                )
                incomplete_args.append(command_arg)
                self.token_ipos_to_arg_map[arg_name_ipos] = command_arg
                arg_name_ipos = None
                arg_name_value = None

            # TODO: TODO_51_14_50_19: ensure `tangent_token` always exists
            #       This may simplify this condition:
            # Post-process `tangent_arg`:
            if (
                self.parsed_ctx.server_action is ServerAction.ProposeArgValues
                and
                self.parsed_ctx.tan_token_ipos >= 0
                and
                self.parsed_ctx.tan_token_ipos in self.token_ipos_to_arg_map
            ):
                tangent_arg = self.token_ipos_to_arg_map[self.parsed_ctx.tan_token_ipos]
                token_bucket_ipos = self.token_ipos_to_token_bucket_map[self.parsed_ctx.tan_token_ipos]
                # FS_23_62_89_43 `tangent_token` completion:
                # `ServerAction.ProposeArgValues` excludes tangent token from consumption because
                # this action is not supposed to consume that tangent token (and complete it instead):
                if isinstance(tangent_arg, ArgCommandIncomplete):
                    # `incomplete_arg` is not consumed - no need to remove:
                    pass
                elif isinstance(tangent_arg, ArgCommandValueDictated):
                    dictated_arg: ArgCommandValueDictated = tangent_arg
                    # TODO: TODO_66_09_41_16: clarify command line processing
                    #       Use helper functions which ensure consistency
                    #       (like in this case, if one is removed, another is removed)
                    self.remaining_dictated_args_per_bucket[token_bucket_ipos].remove(dictated_arg)
                    for token_ipos in dictated_arg.get_arg_tokens():
                        self.remaining_token_buckets[token_bucket_ipos].remove(token_ipos)
                        self.included_token_buckets[token_bucket_ipos].remove(token_ipos)
                        self.excluded_tokens.append(token_ipos)
                elif isinstance(tangent_arg, ArgCommandValueOffered):
                    offered_arg: ArgCommandValueOffered = tangent_arg
                    # TODO: TODO_66_09_41_16: clarify command line processing
                    #       Use helper functions which ensure consistency
                    #       (like in this case, if one is removed, another is removed)
                    self.remaining_offered_args_per_bucket[token_bucket_ipos].remove(offered_arg)
                    self.remaining_token_buckets[token_bucket_ipos].remove(offered_arg.get_arg_token())
                    self.included_token_buckets[token_bucket_ipos].remove(offered_arg.get_arg_token())
                    self.excluded_tokens.append(offered_arg.get_arg_token())
                else:
                    raise TypeError(f"unhandled argument type: {type(tangent_arg)}")
            else:
                # FS_23_62_89_43 `tangent_token` completion:
                # Set all tokens for consumption (including tangent token) in case of
                # `ServerAction.DescribeLineArgs` and `ServerAction.RelayLineArgs`.
                pass

    def remaining_token_ipos_list(
        self,
    ) -> list[int]:
        return token_buckets_to_token_ipos_list(self.remaining_token_buckets)

    def consumed_token_ipos_list(
        self,
    ) -> list[int]:
        return token_buckets_to_token_ipos_list(self.consumed_token_buckets)

    def alloc_searchable_container(
        self,
        base_container_ipos: int,
        func_param_container_offset: int,
        search_control: SearchControl,
    ):
        if (base_container_ipos + func_param_container_offset + 1) >= len(self.envelope_containers):
            envelope_container = EnvelopeContainer(search_control)
            self.envelope_containers.append(envelope_container)

    def is_last_container(self) -> bool:
        return self.curr_container_ipos + 1 == len(self.envelope_containers)

    def is_func_found(self) -> bool:
        return (
            self.curr_container_ipos >= function_container_ipos_
            and
            self.envelope_containers[function_container_ipos_].found_count == 1
        )

    def query_prop_values(self):
        if not self.curr_container:
            return

        ElapsedTime.measure(f"begin_query_envelopes: {self.curr_container.search_control.collection_name}")
        query_dict = populate_query_dict(self.curr_container)
        query_result: QueryResult = self.query_engine.query_prop_values(
            query_dict,
            self.curr_container.search_control,
            self.curr_container.assigned_prop_name_to_prop_value,
        )
        self.curr_container.remaining_prop_name_to_prop_value = query_result.remaining_prop_name_to_prop_value
        self.curr_container.data_envelopes = query_result.data_envelopes
        self.curr_container.found_count = query_result.found_count
        ElapsedTime.measure(f"end_query_envelopes: {query_dict} {self.curr_container.found_count}")

    def consume_args(
        self,
        interp_n: int
    ):
        while True:

            # FS_44_36_84_88 consume args one by one:
            while True:
                # Query envelope values only - they will be used for consumption of command line args:
                ElapsedTime.measure(f"[i={interp_n}]: before_entry_query: {self.curr_interp}")
                self.query_prop_values()
                # Reset to False as we just executed new query:
                arg_was_consumed = False

                # Because each `prop_value` set (per `prop_type`) is treated independently,
                # assignment of `prop_value`-s from args may set combinations
                # which yields no search result in subsequent query.
                # Logically, it is better to consume args one by one and running query after each consumption,
                # but this is not done due to query overhead.
                # Note that Tab-completion and selection (via manual step by human) in separate requests to server and
                # separate `interpret_command` calls is close to that logically better approach.
                if not arg_was_consumed:
                    ElapsedTime.measure(f"[i={interp_n}]: before_consume_dictated_args: {self.curr_interp}")
                    arg_was_consumed = self.curr_interp.consume_dictated_args()
                if not arg_was_consumed:
                    ElapsedTime.measure(f"[i={interp_n}]: before_consume_offered_args: {self.curr_interp}")
                    arg_was_consumed = self.curr_interp.consume_offered_args()

                if arg_was_consumed:
                    if self.curr_interp.consumes_args_at_once():
                        # No known interp consuming at once depends on query - do not query, simply exit loop:
                        break
                    else:
                        # Run next cycle to see if one more can be consumed:
                        pass
                else:
                    break

            query_changed = False

            if self.curr_container:
                # Set implicit values (so that applying defaults knows what they are):
                self.curr_container.populate_implicit_arg_values()

            if self.curr_interp.has_fill_control():
                # FS_72_53_55_13: options before defaults
                # Describing args will need to show options except default - query values before applying defaults:
                ElapsedTime.measure(f"[i={interp_n}]: before_query_without_defaults: {self.curr_interp}")
                # This step may be optimized away and is only needed to detect non-default options for
                # `ServerAction.DescribeLineArgs` but we perform it for all requests to
                # ensure they have common view in case of possible data issues
                # (e.g. TODO_39_25_11_76: `data_envelope`-s with missing props).
                self.query_prop_values()
                # Reset to False as we just executed new query:
                query_changed = False

                self._save_potentially_hidden_by_defaults()

                # Apply defaults (they may apply more than single value at a time):
                query_changed = (
                    self.curr_interp.delegate_fill_control()
                    or
                    query_changed
                )
                self._leave_only_hidden_by_defaults()

            # Query envelopes after all implicit and default values assigned:
            # TODO: We could probably select whether to query only envelopes or
            #       query their values depending on `ServerAction`.
            #       But init of next envelope depends on prev envelope found.
            if query_changed:
                if self.curr_interp.consumes_args_at_once():
                    # No known interp consuming at once depends on query - do not query, simply exit loop:
                    break
                else:
                    pass
            else:
                break

    # TODO: Move all dynamic and non-serializable objects into `InterpRuntime` (or something like that) together with this loop:
    def interpret_command(
        self,
        first_interp_factory_id: str,
    ) -> None:
        """
        Implements FS_55_57_45_04 enum selector version of FS_15_79_76_85 line processor.

        Start with initial interpreter and continue until curr interpreter returns no more next interpreter.

        TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor.
        """

        interp_n: int = 0

        # FS_01_89_09_24 interp tree:
        # start of tracking interp tree abs path while processing command
        # (to pass it to the next interp factory to create next interp):
        self.interp_tree_abs_path = tuple([])

        self.curr_interp = self.create_next_interp(first_interp_factory_id)
        while True:
            interp_n += 1

            self.consume_args(interp_n)

            ElapsedTime.measure(f"[i={interp_n}]: before_try_iterate: {self.curr_interp}")
            interp_step: InterpStep = self.curr_interp.try_iterate()
            ElapsedTime.measure(f"[i={interp_n}]: after_try_iterate: {self.curr_interp}: {interp_step}")
            if interp_step is InterpStep.NextEnvelope:
                continue
            elif interp_step is InterpStep.StopAll:
                ElapsedTime.measure(
                    f"[i={interp_n}]: before_contribute_to_completion: {self.curr_interp}"
                )
                self._contribute_to_completion()
                ElapsedTime.measure(
                    f"[i={interp_n}]: after_contribute_to_completion: {self.curr_interp}"
                )
                return
            elif interp_step is InterpStep.NextInterp:
                ElapsedTime.measure(
                    f"[i={interp_n}]: before_contribute_to_completion: {self.curr_interp}"
                )
                self._contribute_to_completion()
                ElapsedTime.measure(
                    f"[i={interp_n}]: after_contribute_to_completion: {self.curr_interp}"
                )
                self.prev_interp = self.curr_interp
                next_interp = self.curr_interp.next_interp()
                if next_interp:
                    self.curr_interp = next_interp
                else:
                    return
            else:
                raise RuntimeError(interp_step)

    def _save_potentially_hidden_by_defaults(self):
        """
        Save all `remaining_prop_name_to_prop_value` before applying defaults.

        These `filled_prop_values_hidden_by_defaults` are subsequently
        filtered out in `_leave_only_hidden_by_defaults`.
        """
        if not self.curr_container:
            return
        for remaining_prop_name in self.curr_container.remaining_prop_name_to_prop_value.keys():
            self.curr_container.filled_prop_values_hidden_by_defaults[
                remaining_prop_name
            ] = deepcopy(self.curr_container.remaining_prop_name_to_prop_value[remaining_prop_name])

    def _leave_only_hidden_by_defaults(self):
        """
        Weed out those `assigned_prop_name_to_prop_value` from `filled_prop_values_hidden_by_defaults`
        which were set by applying defaults.

        Use `filled_prop_values_hidden_by_defaults` saved before applying defaults
        (in `_save_potentially_hidden_by_defaults` for all `remaining_prop_name_to_prop_value`)
        and remove those which were not hidden by defaults
        to yield only those hidden by defaults.
        """
        if not self.curr_container:
            return
        types_not_hidden_by_defaults = []
        for prop_name_potentially_hidden_by_defaults in self.curr_container.filled_prop_values_hidden_by_defaults.keys():
            if prop_name_potentially_hidden_by_defaults in self.curr_container.assigned_prop_name_to_prop_value:
                if (
                    self.curr_container.assigned_prop_name_to_prop_value[
                        prop_name_potentially_hidden_by_defaults
                    ].value_source
                    is not
                    ValueSource.default_value
                ):
                    types_not_hidden_by_defaults.append(prop_name_potentially_hidden_by_defaults)
            else:
                types_not_hidden_by_defaults.append(prop_name_potentially_hidden_by_defaults)
        # Delete items outside the previous loop:
        for prop_name in types_not_hidden_by_defaults:
            del self.curr_container.filled_prop_values_hidden_by_defaults[prop_name]

    def _contribute_to_completion(self):
        # Each one in the chain of interpreters hava a chance to suggest completion values (contribute):
        self.curr_interp.propose_arg_completion()

    def propose_arg_values(self) -> list[str]:
        """
        FS_71_87_33_52: remove `help_hint` (after first space) if there is only one option
        """
        if len(self.comp_suggestions) == 1:
            first_space_ipos = self.comp_suggestions[0].find(' ')
            if first_space_ipos >= 0:
                self.comp_suggestions[0] = self.comp_suggestions[0][:first_space_ipos]
        return self.comp_suggestions

    def create_next_interp(
        self,
        next_interp_factory_id: str,
    ) -> "AbstractInterp":
        interp_factory: "AbstractInterpFactory" = self.interp_factories[next_interp_factory_id]
        return interp_factory.create_interp(
            self,
        )

    def print_debug(self, end_str: str = "\n") -> None:
        if not self.parsed_ctx.is_debug_enabled:
            return
        self.parsed_ctx.print_debug("")
        eprint(TermColor.debug_output.value, end = "")
        eprint(f"comp_suggestions: {self.comp_suggestions}", end = " ")
        eprint(TermColor.reset_style.value, end = end_str)



from __future__ import annotations

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialFunc import SpecialFunc
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)


class DelegatorEcho(DelegatorSingleFuncAbstract):
    """
    Implements FS_43_50_57_71 `echo_args` func.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        func_envelopes = [{
            instance_data_: {
                func_id_: SpecialFunc.func_id_echo_args.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
            ReservedPropName.help_hint.name: (
                f"Print command line args `{InvocationInput.__name__}`"
            ),
            ReservedPropName.func_state.name: FuncState.fs_beta.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_echo_args.name,
        }]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                self.plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        print(" ".join(invocation_input.all_tokens))



from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_client.client_command_remote.exception_utils import (
    print_full_stack_trace,
    ServerResponseError,
)
from argrelay_app_client.relay_client.ClientCommandAbstract import ClientCommandAbstract
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.enum_desc.ProcRole import ProcRole


class ClientCommandRemoteAbstract(ClientCommandAbstract):

    def __init__(
        self,
        call_ctx: CallContext,
        proc_role: ProcRole,
    ):
        super().__init__(
            call_ctx,
        )
        self.proc_role = proc_role

    @staticmethod
    def raise_error(
        status_code: int,
    ):
        raise ServerResponseError(f"server response status_code: {status_code}")

    def execute_command(
        self,
    ):
        self._execute_single_call()

    def _execute_single_call(
        self,
    ):
        """
        Basic implementation of remote call to a single server.

        See `ClientCommandRemoteWorkerAbstract._execute_multiple_calls` for alternative.
        """
        try:
            self._execute_remote_call()
        except (
            ConnectionError,
            ConnectionRefusedError,
        ) as e:
            self._handle_exception_with_exit_code(
                False,
                e,
                ClientExitCode.ConnectionError.value,
            )
        except ServerResponseError as e:
            self._handle_exception_with_exit_code(
                True,
                e,
                ClientExitCode.ServerError.value,
            )

    def _handle_exception_with_exit_code(
        self,
        print_stack_trace_on_exit,
        exception_obj,
        exit_code,
    ):
        """
        Raise exception or exit with exit code depending on `ProcRole`.
        """
        if (
            self.proc_role.is_worker_proc
            and
            # FS_36_17_84_44 check_env: do not exit:
            self.proc_role is not ProcRole.CheckEnvWorker
        ):
            if print_stack_trace_on_exit:
                print_full_stack_trace(exception_obj)
            # Tell parent what happened (let parent talk the rest):
            exit(exit_code)
        else:
            raise exception_obj

    def _execute_remote_call(
        self,
    ):
        pass



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_lib_root.enum_desc.DistinctValuesQuery import DistinctValuesQuery
from argrelay_schema_config_server.runtime_data_server_app.MongoClientConfig import MongoClientConfig
from argrelay_schema_config_server.runtime_data_server_app.MongoServerConfig import MongoServerConfig


@dataclass
class MongoConfig:
    use_mongomock: bool = field()
    """
    See :class:`MongoConfigSchema.use_mongomock`.
    """

    distinct_values_query: DistinctValuesQuery = field()
    """
    See `DistinctValuesQuery`.
    """

    mongo_client: MongoClientConfig = field()

    mongo_server: MongoServerConfig = field()



from __future__ import annotations

from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorNoopBase import DelegatorNoopBase


class DelegatorNoopEmpty(DelegatorNoopBase):
    """
    The purpose of this delegator is NOT to have implementation of any function.
    """



from __future__ import annotations

import os.path
from typing import Union

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory
from argrelay_lib_root.misc_helper_common import get_argrelay_dir
from argrelay_lib_server_plugin_demo.demo_git.git_utils import (
    get_unversioned_file_list,
    is_git_repo,
)


class PluginCheckEnvArgrelayUnversionedSymlinks(PluginCheckEnvAbstract):
    """
    Plugin reports unversioned and unignored symlinks.

    Some common files are normally installed as symlinks rather than a copy.

    It is likely that those unversioned and unignored symlinks are leftover of upgrades
    (via `@/exe/bootstrap_env.bash`) from one version to another because
    such upgrade procedures are NOT complicated by tracking what was in
    multitude of old versions to be changed on upgrade to multiple of new versions.
    """

    field_name = "unversioned_symlink"

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        argrelay_dir = get_argrelay_dir()

        if is_git_repo(argrelay_dir):
            try:
                unversioned_file_list: list[str] = get_unversioned_file_list(argrelay_dir)
            except ValueError:
                unversioned_file_list = None
        else:
            unversioned_file_list = None

        if unversioned_file_list is None:
            return [CheckEnvResult(
                result_category = ResultCategory.ExecutionFailure,
                result_key = self.field_name,
                result_value = None,
                result_message = "Failed to retrieve unversioned files from git repo.",
            )]
        else:
            results_per_file: list[CheckEnvResult] = []
            for unversioned_file in unversioned_file_list:
                full_path = os.path.join(argrelay_dir, unversioned_file)
                if os.path.islink(full_path):
                    results_per_file.append(CheckEnvResult(
                        result_category = ResultCategory.VerificationWarning,
                        result_key = f"{self.field_name}[{len(results_per_file)}]",
                        result_value = full_path,
                        result_message = "Running `@/exe/bootstrap_env.bash` may leave unversioned unignored symlinks - review and decide to remove or ignore them manually.",
                    ))

            if len(results_per_file) > 0:
                return results_per_file
            else:
                return [CheckEnvResult(
                    result_category = ResultCategory.VerificationSuccess,
                    result_key = self.field_name,
                    result_value = None,
                    result_message = "Clean: no unversioned unignored symlinks.",
                )]



from argrelay_api_server_cli.schema_request.CallContextSchema import call_context_desc
from argrelay_api_server_cli.schema_response.AssignedValueSchema import assigned_value_desc
from argrelay_api_server_cli.schema_response.EnvelopeContainerSchema import envelope_container_desc
from argrelay_api_server_cli.schema_response.InterpResultSchema import interp_result_desc
from argrelay_api_server_cli.server_spec.server_data_schema import get_schema_definitions
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import data_envelope_desc
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import search_control_desc

spec_data = {
    "summary": "Describe search results from `prop_value`-s of the given command line and cursor position",
    "consumes": [
        "application/json",
    ],
    "produces": [
        "application/json",
    ],
    "parameters": [
        {
            "name": call_context_desc.ref_name,
            "in": "body",
            "required": True,
            "schema": {
                "$ref": "#/definitions/" + call_context_desc.ref_name,
            },
        },
    ],
    "responses": {
        "200": {
            "description": "Described search results",
            "schema": {
                "$ref": "#/definitions/" + interp_result_desc.ref_name,
            },
            "examples": {
                "application/json": interp_result_desc.dict_example,
            },
        },
    },
    "definitions": get_schema_definitions([
        # TODO: provide `call_context_desc` with example of expected `comp_type` for `describe_line_args`:
        call_context_desc.ref_name,
        interp_result_desc.ref_name,
        envelope_container_desc.ref_name,
        data_envelope_desc.ref_name,
        assigned_value_desc.ref_name,
        search_control_desc.ref_name,
    ]),
}



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_lib_root.schema_config.ConnectionConfig import ConnectionConfig


@dataclass(frozen = True)
class ClientConfig:
    __comment__: str = field()
    use_local_requests: bool = field()
    optimize_completion_request: bool = field()
    redundant_servers: list[ConnectionConfig] = field()
    show_pending_spinner: bool = field()
    spinless_sleep_sec: float = field()



from __future__ import annotations

import re
from typing import Union

import argrelay
from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory
from argrelay_lib_root.misc_helper_common import get_argrelay_dir


class PluginCheckEnvArgrelayLocalVersionMismatch(PluginCheckEnvAbstract):
    """
    Plugin reports mismatch between
    *   `argrelay.__version__`
    *   version in `@/conf/env_packages.txt`
    """

    # TODO: TODO_69_59_78_78: register known files as enum with metadata:
    file_rel_path = "conf/env_packages.txt"
    bootstrap_rel_path = "exe/bootstrap_env.bash"
    result_key = "env_packages_version"

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        module_version = argrelay.__version__
        env_packages_version = self.get_env_packages_version()
        if env_packages_version is None:
            return [CheckEnvResult(
                result_category = ResultCategory.VerificationWarning,
                result_key = self.result_key,
                result_value = env_packages_version,
                result_message = f"File `@/{self.file_rel_path}` does not contain `argrelay` package => is `argrelay` installed in editable mode?",
            )]
        else:
            if env_packages_version == module_version:
                return [CheckEnvResult(
                    result_category = ResultCategory.VerificationSuccess,
                    result_key = self.result_key,
                    result_value = env_packages_version,
                    result_message = f"It matches `argrelay_module_version`",
                )]
            else:
                return [CheckEnvResult(
                    result_category = ResultCategory.VerificationFailure,
                    result_key = self.result_key,
                    result_value = env_packages_version,
                    result_message = f"It does not match `argrelay_module_version` => re-base and re-run `@/{self.bootstrap_rel_path}`",
                )]

    def get_env_packages_version(
        self,
    ):
        file_content = open(f"{get_argrelay_dir()}/{self.file_rel_path}", "rt").read()
        version_regex = r"^argrelay==(.*)$"
        regex_match = re.search(version_regex, file_content, re.M)
        if regex_match:
            return regex_match.group(1).strip()
        else:
            return None



from enum import (
    auto,
    Enum,
)


class GitRepoEnvelopeClass(Enum):
    class_git_repo = auto()
    class_git_tag = auto()
    class_git_commit = auto()

    def __str__(self):
        return self.name



"""
"""
import os
from contextlib import contextmanager
from inspect import (
    currentframe,
    getframeinfo,
)
from pathlib import PurePath
from types import ModuleType

test_data_ = "test_data"

# Enable/disable breakpoints programmatically:
__is_breakpoint_enabled = False


def enable_breakpoint():
    global __is_breakpoint_enabled
    __is_breakpoint_enabled = True


def disable_breakpoint():
    global __is_breakpoint_enabled
    __is_breakpoint_enabled = False


def is_breakpoint_enabled() -> bool:
    return __is_breakpoint_enabled


def line_no() -> int:
    """
    Get source line in the caller frame
    """
    return line_no_l_0()


def line_no_from_ctor() -> int:
    """
    Same as `line_no` but used inside `TestCase` ctor (behind one more stack frame).
    """
    return line_no_l_1()


def line_no_l_0() -> int:
    """
    Line no from frame level 0 = callee frame = curr frame (from call site POV)
    """
    return getframeinfo(currentframe().f_back).lineno


def line_no_l_1() -> int:
    """
    Line no from frame level 1 = caller frame = prev frame (from call site POV)
    """
    return getframeinfo(currentframe().f_back.f_back).lineno


def file_line() -> str:
    return file_line_l_0()


def file_line_l_0() -> str:
    """
    Similar to `line_no_l_0` but includes file name.
    """
    stack_frame = currentframe().f_back
    return f"{stack_frame.f_code.co_filename}:{stack_frame.f_lineno}"


def file_line_l_1() -> str:
    """
    Similar to `file_line_l_1` but includes file name.
    """
    stack_frame = currentframe().f_back.f_back
    return f"{stack_frame.f_code.co_filename}:{stack_frame.f_lineno}"


def parse_line_and_cpos(test_line: str) -> (str, int):
    """
    Translate test line with pipe as cursor place into command line string and cursor char position (cpos)
    """
    assert test_line.count('|') == 1
    cursor_cpos = test_line.index('|')
    command_line = test_line.replace('|', "")
    return command_line, cursor_cpos


@contextmanager
def change_to_known_repo_path(path_from_repo_root = "./tests"):
    """
    This function changes current dir to known path within repo root.

    This allows any other code accessing files by relative paths rely on the stable path within the repo.
    """

    old_pwd = os.getcwd()
    try:
        # When IDE runs, CWD is somewhere under "tests".
        # Keep climbing up until directory becomes "tests":
        if "tests" in PurePath(os.getcwd()).parts:
            while os.path.basename(os.getcwd()) != "tests":
                os.chdir("..")

        # When `tox` runs, CWD = [repo root]:
        if os.path.basename(os.getcwd()) != "tests":
            os.chdir("tests")

        # Now, when same baseline is set regardless `tox` or IDE, change to repo root:
        os.chdir("..")

        # Desired path:
        os.chdir(path_from_repo_root)
        yield
    finally:
        os.chdir(old_pwd)


def assert_test_func_name_embeds_prod_class_name(
    prod_class: type,
):
    """
    Ensure caller test function name contains given prod class name.
    """
    caller_func_name = currentframe().f_back.f_code.co_name
    simple_prod_class_name = prod_class.__name__.split(".")[-1]
    _assert_test_name_embeds_prod_name(simple_prod_class_name, caller_func_name)


def assert_test_module_name_embeds_prod_class_name(
    prod_class: type,
):
    """
    Ensure caller test module name contains given prod class name.
    """
    caller_frame = currentframe().f_back
    simple_test_module_name = caller_frame.f_globals["__name__"].split(".")[-1]
    simple_prod_class_name = prod_class.__name__.split(".")[-1]
    _assert_test_name_embeds_prod_name(simple_prod_class_name, simple_test_module_name)


def assert_test_module_name_embeds_prod_module_name(
    prod_module: ModuleType,
):
    """
    Ensure caller test module name contains given prod module name.
    """
    caller_frame = currentframe().f_back
    simple_test_module_name = caller_frame.f_globals["__name__"].split(".")[-1]
    simple_prod_module_name = prod_module.__name__.split(".")[-1]
    _assert_test_name_embeds_prod_name(simple_prod_module_name, simple_test_module_name)


def _assert_test_name_embeds_prod_name(
    prod_name: str,
    test_name: str,
):
    """
    This function ensures that names in prod code and test code do not diverge due to refactoring.

    That programmatically establishes relationship between prod code and test code via cross-references.
    The function should not be called directly (with strings which defeats the purpose as strings easily diverge).
    Instead, an appropriate wrapper function should be called with references (e.g. to classes).
    """
    assert prod_name in test_name



from enum import (
    auto,
    Enum,
)


class ResultCategory(Enum):
    """
    Used for implementation of FS_36_17_84_44 `check_env`.
    """

    VerificationSuccess = auto()
    """
    Able to retrieve data to check and the check is successful.
    """

    VerificationWarning = auto()
    """
    Able to retrieve data to check, the result of the check is successful but not not absolutely clean.
    """

    VerificationFailure = auto()
    """
    Able to retrieve data to check, but the result of the check is not successful.
    """

    ServerOffline = auto()
    """
    Unable to retrieve data to check for the check which required server online.
    """

    ExecutionFailure = auto()
    """
    Unable to retrieve data to check.
    """



"""
"""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    Union,
)

from argrelay_api_plugin_server_abstract.ConfiguratorAbstract import ConfiguratorAbstract
from argrelay_api_server_cli.schema_response.AssignedValue import AssignedValue
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.enum_desc.SpecialChar import SpecialChar
from argrelay_lib_root.enum_desc.ValueSource import ValueSource
from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorError import DelegatorError
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_code_,
    error_delegator_custom_data_desc,
    error_message_,
)
from argrelay_schema_config_server.runtime_data_server_plugin.PluginConfig import PluginConfig


def set_default_to(
    prop_name,
    prop_value,
    envelope_container,
) -> bool:
    """
    Assign a default `prop_value` to the given `prop_name`:
    *   Ensure that this `prop_name` is valid for `envelope_container`.
    *   Ensure that this `prop_value` is actually one of the remaining ones.
    """

    if prop_name in envelope_container.search_control.prop_name_to_arg_name_dict:
        if prop_name not in envelope_container.assigned_prop_name_to_prop_value:
            if prop_name in envelope_container.remaining_prop_name_to_prop_value:
                if prop_value in envelope_container.remaining_prop_name_to_prop_value[prop_name]:
                    del envelope_container.remaining_prop_name_to_prop_value[prop_name]
                    envelope_container.assigned_prop_name_to_prop_value[prop_name] = AssignedValue(
                        prop_value,
                        ValueSource.default_value,
                    )
                    return True
        else:
            # Setting `ValueSource.default_value`: it cannot override any, right? No point to handle assigned case:
            pass
    return False


def get_config_value_once(
    server_configurators,
    value_getter: Callable[[ConfiguratorAbstract], Any],
    default_value: Any,
) -> Any:
    config_value: Union[Any, None] = None
    server_configurator: ConfiguratorAbstract
    for server_configurator in server_configurators.values():
        if config_value is None:
            config_value = value_getter(server_configurator)
        else:
            if value_getter(server_configurator) is not None:
                # Only one `PluginType.ConfiguratorPlugin` providing
                # same type of value is supported to avoid confusion:
                raise RuntimeError

    if config_value is None:
        config_value = default_value
    return config_value


def redirect_to_error(
    interp_ctx,
    plugin_config: PluginConfig,
    error_message,
    error_code,
):
    # Redirect to `DelegatorError`:
    # TODO: TODO_62_75_33_41: Do not hardcode plugin instance id (instance of `DelegatorError`):
    delegator_plugin_instance_id = f"{DelegatorError.__name__}.default"
    custom_plugin_data = {
        error_message_: error_message,
        error_code_: error_code,
    }
    error_delegator_custom_data_desc.validate_dict(custom_plugin_data)
    invocation_input = InvocationInput.with_interp_context(
        interp_ctx,
        delegator_plugin_entry = plugin_config.server_plugin_instances[delegator_plugin_instance_id],
        custom_plugin_data = custom_plugin_data,
    )
    return invocation_input


def redirect_to_not_disambiguated_error(
    interp_ctx,
    plugin_config,
    envelope_class,
):
    return redirect_to_error(
        interp_ctx,
        plugin_config,
        f"ERROR: `envelope_class` [{envelope_class}] is not fully qualified (not disambiguated)",
        ClientExitCode.GeneralError.value,
    )


def clean_prop_value(
    prop_value: str,
) -> str:
    if prop_value == SpecialChar.NoPropValue.value:
        return ""
    else:
        return prop_value



from typing import Union

from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory


class CheckEnvResult:
    """
    Result data from plugin Implements FS_36_17_84_44 `check_env`.
    """

    def __init__(
        self,
        result_category: ResultCategory,
        result_key: Union[str, None],
        result_value: Union[str, None],
        result_message: Union[str, None],
    ):
        self.result_category: ResultCategory = result_category
        self.result_key: Union[str, None] = result_key
        self.result_value: Union[str, None] = result_value
        self.result_message: Union[str, None] = result_message



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)
from typing import Collection

from argrelay_lib_root.misc_helper_common import eprint


@dataclass
class ProgressTracker:
    """
    State and stats to track `data_envelope`-s loading progress.
    """
    envelope_per_col_i: int = field(default = 0)
    envelope_per_col_n: int = field(default = 0)
    base_total_envelope_i: int = field(default = 0)
    total_envelope_i: int = field(default = 0)
    total_envelope_n: int = field(default = 0)

    # Number of envelopes loaded per `collection_name`:
    collection_sizes: dict[str, int] = field(default_factory = dict)

    # TODO: Provide a way to exclude some collection_name and envelope_classes from
    #       requirement of being part of `search_control`.
    #       For example, `ReservedEnvelopeClass.class_help` is used internally (may not be used by `search_control`).
    unused_index_props_per_collection: dict[str, set[str]] = field(default_factory = lambda: {})

    dangling_search_props_per_collection: dict[str, set[str]] = field(default_factory = lambda: {})

    def track_unused_index_props_per_collection_unused_by_search_control(
        self,
        collection_name: str,
        index_props: Collection[str],
    ):
        """
        TODO: deduplicate with `track_dangling_search_props_per_collection_undefined_by_index_model`
        """
        assert len(index_props) > 0
        self.unused_index_props_per_collection.setdefault(
            collection_name,
            set(),
        ).update(index_props)

        prop_names = " ".join(index_props)
        eprint(
            f"`index_model` components not used by any `search_control`: "
            f"`collection_name` [{collection_name}] "
            f"`index_props`: {prop_names} "
        )

    def track_dangling_search_props_per_collection_undefined_by_index_model(
        self,
        collection_name: str,
        search_props: Collection[str],
    ):
        """
        TODO: deduplicate with `track_unused_index_props_per_collection_unused_by_search_control`
        """
        assert len(search_props) > 0
        self.dangling_search_props_per_collection.setdefault(
            collection_name,
            set(),
        ).update(search_props)

        prop_names = " ".join(search_props)
        error_message = (
            f"`collection_name` [{collection_name}] "
            f"`search_props`: {prop_names} "
        )
        if self.collection_sizes.get(collection_name, 0) == 0:
            eprint(
                f"`search_control` components not in `index_model` (but ignoring as collection is empty now): " +
                error_message
            )
        else:
            raise ValueError(
                f"`search_control` components not in `index_model`: " +
                error_message
            )

    def report_collection_sizes(
        self,
    ):
        for collection_name, collection_size in self.collection_sizes.items():
            if collection_size != 0:
                eprint(f"`collection_name` [{collection_name}] size: {collection_size}")
        for collection_name, collection_size in self.collection_sizes.items():
            if collection_size == 0:
                eprint(f"`index_model` declares `collection_name` [{collection_name}] but it is empty")

    def _assert_intermediate_progress(
        self,
    ):
        assert self.envelope_per_col_i <= self.envelope_per_col_n
        assert self.envelope_per_col_n <= self.total_envelope_n
        assert self.total_envelope_i <= self.total_envelope_n

    def _assert_collection_final_progress(
        self,
    ):
        self._assert_intermediate_progress()
        assert self.envelope_per_col_i == self.envelope_per_col_n

    def _assert_total_final_progress(
        self,
    ):
        self._assert_collection_final_progress()
        if self.total_envelope_i != self.total_envelope_n:
            raise AssertionError(
                f"`total_envelope_i` [{self.total_envelope_i}] == "
                f"`total_envelope_n` [{self.total_envelope_n}] "
                f"(not true)"
            )

    def track_collection_size_increment(
        self,
        collection_name: str,
        increment_size: int,
    ):
        """
        Track collection size increment by each store step (after each loader plugin).

        NOTE: It is assumed each loader appends only (does not override existing `data_envelope`-s).
        """
        curr_size = self.collection_sizes.setdefault(collection_name, 0)
        self.collection_sizes[collection_name] = curr_size + increment_size

    def track_collection_indexing_start(
        self,
        collection_name: str,
        increment_size: int,
    ):
        self.base_total_envelope_i: int = self.total_envelope_i
        self.envelope_per_col_i = 0
        self.envelope_per_col_n = increment_size
        self.log_collection_indexing_progress(collection_name)

        assert increment_size <= self.collection_sizes[collection_name]

    def track_collection_indexing_stop(
        self,
        collection_name: str,
    ):
        assert self.envelope_per_col_i == self.total_envelope_i - self.base_total_envelope_i
        self.log_collection_indexing_progress(collection_name)

    def track_collection_indexing_increment(
        self,
        collection_name: str,
    ):
        if self.total_envelope_i > 0 and self.total_envelope_i % 1_000 == 0:
            self.log_collection_indexing_progress(collection_name)
        self.total_envelope_i += 1
        self.envelope_per_col_i += 1

    def log_collection_indexing_progress(
        self,
        collection_name: str,
    ):
        try:
            self._assert_intermediate_progress()
        finally:
            eprint(
                f"collection: {collection_name}: indexed envelopes: "
                f"{self.envelope_per_col_i}/{self.envelope_per_col_n} "
                f"{self.total_envelope_i}/{self.total_envelope_n} "
                "..."
            )

    def track_total_validation_start(
        self,
    ):
        self.total_envelope_i = 0

    def track_total_validation_stop(
        self,
    ):
        self._assert_total_final_progress()
        self.report_collection_sizes()

    def track_collection_validation_start(
        self,
        collection_name: str,
    ):
        eprint(f"collection to validate: {collection_name}")
        self.envelope_per_col_n = self.collection_sizes.get(collection_name, 0)
        self.envelope_per_col_i = 0
        self.log_collection_validation_progress(collection_name)

    def track_collection_validation_stop(
        self,
        collection_name: str,
    ):
        self.log_collection_validation_progress(
            collection_name,
        )
        self._assert_collection_final_progress()

    def track_collection_validation_increment(
        self,
        collection_name: str,
    ):
        self.envelope_per_col_i += 1
        self.total_envelope_i += 1

        if self.total_envelope_i > 0 and self.total_envelope_i % 1_000 == 0:
            self.log_collection_validation_progress(
                collection_name,
            )

    def log_collection_validation_progress(
        self,
        mongo_collection: str,
    ):
        try:
            self._assert_intermediate_progress()
        finally:
            eprint(
                f"collection: {mongo_collection}: validated envelopes: "
                f"{self.envelope_per_col_i}/{self.envelope_per_col_n} "
                f"{self.total_envelope_i}/{self.total_envelope_n} "
                f"..."
            )



from __future__ import annotations

from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_demo.demo_service.DelegatorServiceBase import DelegatorServiceBase
from argrelay_lib_server_plugin_demo.demo_service.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay_lib_server_plugin_demo.demo_service.ServicePropName import ServicePropName
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import populate_search_control


def get_host_search_control(
) -> dict:
    return populate_search_control(
        ServiceEnvelopeClass.class_host.name,
        {
            ReservedPropName.envelope_class.name: ServiceEnvelopeClass.class_host.name,
        },
        [
            # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
            {"class": ReservedPropName.envelope_class.name},

            # class_cluster:
            {"code": ServicePropName.code_maturity.name},
            {"stage": ServicePropName.flow_stage.name},
            {"region": ServicePropName.geo_region.name},
            {"cluster": ServicePropName.cluster_name.name},
            # class_host:
            {"host": ServicePropName.host_name.name},
            # ---
            {"status": ServicePropName.live_status.name},
            {"ip": ServicePropName.ip_address.name},
        ],
    )


class DelegatorServiceHostBase(DelegatorServiceBase):
    """
    Provide base functionality for funcs working with `ServiceEnvelopeClass.class_host`.
    """



from __future__ import annotations

import os

from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_client.client_command_remote.ClientCommandRemoteAbstract import ClientCommandRemoteAbstract
from argrelay_app_client.client_command_remote.exception_utils import ServerResponseError
from argrelay_app_client.client_pipeline import BytesSrcAbstract
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.enum_desc.ProcRole import ProcRole
from argrelay_lib_root.enum_desc.TopDir import TopDir
from argrelay_lib_root.misc_helper_common import get_argrelay_dir
from argrelay_lib_root.schema_config.ConnectionConfig import ConnectionConfig

server_index_file_name = "argrelay_client.server_index"
random_file = "/dev/random"


def get_var_dir_path() -> str:
    return str(os.path.join(get_argrelay_dir(), TopDir.var_dir.value))


def get_server_index_file_path() -> str:
    return str(os.path.join(get_var_dir_path(), server_index_file_name))


def load_server_index(
) -> int:
    server_index_file_path = get_server_index_file_path()
    if os.path.isfile(server_index_file_path):
        with open(server_index_file_path, "r") as open_file:
            return int(open_file.read())
    else:
        with open(random_file, "rb") as open_file:
            return int.from_bytes(open_file.read(1), "little")


def store_server_index(
    curr_server_index: int,
) -> None:
    os.makedirs(
        get_var_dir_path(),
        exist_ok = True,
    )
    server_index_file_path = get_server_index_file_path()
    with open(server_index_file_path, "w") as open_file:
        open_file.write(str(curr_server_index))


class ClientCommandRemoteWorkerAbstract(ClientCommandRemoteAbstract):

    def __init__(
        self,
        call_ctx: CallContext,
        proc_role: ProcRole,
        redundant_servers: list[ConnectionConfig],
        server_index: int,
        bytes_src: BytesSrcAbstract,
    ):
        super().__init__(
            call_ctx,
            proc_role,
        )
        self.redundant_servers: list[ConnectionConfig] = redundant_servers
        self.bytes_src: BytesSrcAbstract = bytes_src
        if server_index < 0:
            self.curr_connection_config: ConnectionConfig = self.redundant_servers[0]
            self.use_round_robin = True
        else:
            self.curr_connection_config: ConnectionConfig = self.redundant_servers[server_index]
            self.use_round_robin = False

        self.server_index_file_path: str = get_server_index_file_path()

    def execute_command(
        self,
    ):
        if not self.use_round_robin:
            self._execute_single_call()
        else:
            self._execute_multiple_calls()

    def _execute_multiple_calls(
        self,
    ):
        """
        Implements FS_93_18_57_91 client fail over.

        For basic implementation (no round-robin) see base `ClientCommandRemoteAbstract` class.
        """
        connections_count = len(self.redundant_servers)
        init_server_index = load_server_index()
        for curr_connection_offset in range(connections_count):
            curr_server_index = (init_server_index + curr_connection_offset) % connections_count
            self.curr_connection_config = self.redundant_servers[curr_server_index]

            try:
                self._execute_remote_call()
                store_server_index(curr_server_index)
                return
            except (
                ConnectionError,
                ConnectionRefusedError,
            ):
                continue
            except ServerResponseError as e:
                self._handle_exception_with_exit_code(
                    True,
                    e,
                    ClientExitCode.ServerError.value,
                )

        self._handle_exception_with_exit_code(
            True,
            ConnectionError(f"Unable to connect to any of [{connections_count}] configured connections"),
            ClientExitCode.ConnectionError.value,
        )



from __future__ import annotations

from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_demo.demo_service.DelegatorServiceBase import DelegatorServiceBase
from argrelay_lib_server_plugin_demo.demo_service.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay_lib_server_plugin_demo.demo_service.ServicePropName import ServicePropName
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import (
    populate_search_control,
)


def get_cluster_search_control(
) -> dict:
    return populate_search_control(
        ServiceEnvelopeClass.class_cluster.name,
        {
            ReservedPropName.envelope_class.name: ServiceEnvelopeClass.class_cluster.name,
        },
        [
            # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
            {"class": ReservedPropName.envelope_class.name},

            {"code": ServicePropName.code_maturity.name},
            {"stage": ServicePropName.flow_stage.name},
            {"region": ServicePropName.geo_region.name},
            {"cluster": ServicePropName.cluster_name.name},
        ],
    )


class DelegatorServiceClusterBase(DelegatorServiceBase):
    """
    Provide base functionality for funcs working with `ServiceEnvelopeClass.class_cluster`.
    """



from __future__ import annotations

import re
import subprocess
from typing import Union

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory

_common_banner = "Search about GNU Readline (`~/.inputrc`) or `bind` in Bash."
_common_bell_explanation = "Bash sometimes gives terminal \"bell\" instead of showing auto-completion."


class PluginCheckEnvReadlineAbstract(PluginCheckEnvAbstract):
    """
    This plugin checks config of Readline library:
    https://en.wikipedia.org/wiki/GNU_Readline

    Official Bash doc:
    https://www.gnu.org/software/bash/manual/html_node/Readline-Init-File-Syntax.html

    It may be configured via `~/.inputrc` and is used by Bash which can show its settings via `bind -v` command.

    NOTE: Some of these settings may be forced by `@/exe/dev_shell.bash` (set in `@/exe/init_shell_env.bash`),
          but this plugin cannot verify current Bash instance
          (its properties are transient in the memory of its process).
          Instead, it verifies configured properties set for each new Bash instance.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
        #
        prop_name: str,
        #
        if_on_message: str,
        if_on_category: ResultCategory,
        #
        if_off_message: str,
        if_off_category: ResultCategory,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
        )
        #
        self.prop_name = prop_name
        #
        self.if_on_message = if_on_message
        self.if_on_category = if_on_category
        #
        self.if_off_message = if_off_message
        self.if_off_category = if_off_category

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        try:
            bash_proc = subprocess.run(
                args = [
                    "bash",
                    "-c",
                    "bind -v",
                ],
                capture_output = True,
            )
            assert bash_proc.returncode == 0
            stdout_str = bash_proc.stdout.decode("utf-8")

            prop_value = None
            prop_regex = f"^set\\s*{self.prop_name}\\s*(\\S*)\\s*$"
            for output_line in stdout_str.splitlines():
                regex_match = re.search(prop_regex, output_line, re.M)
                if regex_match:
                    prop_value = regex_match.group(1).strip()
        except:
            prop_value = None

        if "on" == prop_value:
            return [CheckEnvResult(
                result_category = self.if_on_category,
                result_key = self.prop_name,
                result_value = str(prop_value),
                result_message = self.if_on_message,
            )]
        elif "off" == prop_value:
            return [CheckEnvResult(
                result_category = self.if_off_category,
                result_key = self.prop_name,
                result_value = str(prop_value),
                result_message = self.if_off_message,
            )]
        elif prop_value is not None:
            return [CheckEnvResult(
                result_category = ResultCategory.ExecutionFailure,
                result_key = self.prop_name,
                result_value = str(prop_value),
                result_message = f"Unknown value for `{prop_value}` in `bind -v` output.",
            )]
        else:
            return [CheckEnvResult(
                result_category = ResultCategory.ExecutionFailure,
                result_key = self.prop_name,
                result_value = str(prop_value),
                result_message = f"Unable to parse `bind -v` output and find `{prop_value}`.",
            )]


class PluginCheckEnvReadlineShowAllIfAmbiguous(PluginCheckEnvReadlineAbstract):
    """
    See more: https://stackoverflow.com/a/42193784/441652
    """

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        prop_name = "show-all-if-ambiguous"
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
            prop_name = prop_name,
            if_on_message = f"More convenient. {_common_banner}",
            if_on_category = ResultCategory.VerificationSuccess,
            if_off_message = f"Less convenient: if `{prop_name}` is off, {_common_bell_explanation} {_common_banner}",
            if_off_category = ResultCategory.VerificationWarning,
        )


class PluginCheckEnvReadlineShowAllIfUnmodified(PluginCheckEnvReadlineAbstract):
    """
    See more: https://stackoverflow.com/a/42193784/441652
    """

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        prop_name = "show-all-if-unmodified"
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
            prop_name = prop_name,
            if_on_message = f"More convenient. {_common_banner}",
            if_on_category = ResultCategory.VerificationSuccess,
            if_off_message = f"Less convenient: if `{prop_name}` is off, {_common_bell_explanation} {_common_banner}",
            if_off_category = ResultCategory.VerificationWarning,
        )


class PluginCheckEnvReadlineSkipCompletedText(PluginCheckEnvReadlineAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        prop_name = "skip-completed-text"
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
            prop_name = prop_name,
            if_on_message = f"More convenient. {_common_banner}",
            if_on_category = ResultCategory.VerificationSuccess,
            if_off_message = f"Less convenient: if `{prop_name}` is off, inserting completed CLI arg part will duplicate existing one. {_common_banner}",
            if_off_category = ResultCategory.VerificationWarning,
        )



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_lib_root.enum_desc.PluginSide import PluginSide


@dataclass(frozen = True)
class PluginEntry:
    plugin_enabled: bool = field()
    plugin_side: PluginSide = field()
    plugin_module_name: str = field()
    plugin_class_name: str = field()
    plugin_dependencies: list[str] = field()
    plugin_config: dict = field()



from __future__ import annotations

from typing import Union

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_plugin_server_abstract.SchemaConfigDelegatorJumpAbstract import (
    abstract_jump_delegator_config_desc,
    single_func_id_,
)
from argrelay_app_server.composite_forest.CompositeForestExtractor import extract_tree_abs_path_to_interp_id
from argrelay_app_server.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay_app_server.composite_forest.DictTreeWalker import DictTreeWalker
from argrelay_lib_server_plugin_core.plugin_interp.AbstractInterpFactory import AbstractInterp
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig


class DelegatorJumpAbstract(DelegatorSingleFuncAbstract):
    """
    This delegator supports functionality of FS_91_88_07_23 jump tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        # Use configured `single_func_id` to extract all tree paths where it is used:
        extracted_dict = extract_tree_abs_path_to_interp_id(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_config_dict[single_func_id_],
        )
        dict_tree_walker = DictTreeWalker(
            CompositeInfoType.tree_abs_path_to_interp_id,
            extracted_dict,
        )
        # Temporary (reversed) 1-to-N map of id to paths (instead of path to id):
        temporary_id_to_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()

        # Reverse temporary 1-to-N map of id to paths into 1-to-1 map of path to id:
        self.tree_path_to_next_interp_plugin_instance_id: dict[tuple[str, ...], str] = {}
        for interp_factory_instance_id, tree_abs_paths in temporary_id_to_paths.items():
            for tree_abs_path in tree_abs_paths:
                assert tuple(tree_abs_path) not in self.tree_path_to_next_interp_plugin_instance_id
                self.tree_path_to_next_interp_plugin_instance_id[tuple(tree_abs_path)] = interp_factory_instance_id

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return abstract_jump_delegator_config_desc.dict_from_input_dict(plugin_config_dict)

    def run_interp_control(
        self,
        curr_interp: AbstractInterp,
    ) -> Union[None, str]:
        """
        Select next interp for some of the `SpecialFunc`-s:
        *   `SpecialFunc.func_id_intercept_invocation`
        *   `SpecialFunc.func_id_help_hint`
        *   ...

        The next interp selection is a "double jump":
        *   the first "plugin_id" jump here (in delegator) to interp via delegator of the selected func
        *   the second "tree_path" jump there (in jump tree interp) from jump interp to new tree path via jump tree

        This delegator (its `single_func_id`) can be accessed through multiple tree paths.
        Therefore, it should map `interp_tree_abs_path` (how it is accessed) to
        the next `interp_plugin_instance_id` (the single interp).
        """
        if curr_interp.interp_ctx.interp_tree_abs_path in self.tree_path_to_next_interp_plugin_instance_id:
            return self.tree_path_to_next_interp_plugin_instance_id[curr_interp.interp_ctx.interp_tree_abs_path]
        else:
            return None



import time
from datetime import (
    datetime,
    timezone,
)

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_lib_check_env_plugin_core.check_env_utils import format_time_to_relative
from argrelay_lib_check_env_plugin_core.PluginCheckEnvServerResponseValueAbstract import (
    PluginCheckEnvServerResponseValueAbstract,
)
from argrelay_lib_check_env_plugin_core.SchemaPluginCheckEvnServerResponseValueAbstract import (
    field_values_to_command_lines_,
)
from argrelay_lib_root.enum_desc.CheckEnvField import CheckEnvField


class PluginCheckEnvServerResponseValueStartTime(PluginCheckEnvServerResponseValueAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict = plugin_config_dict or {
                field_values_to_command_lines_: {
                    CheckEnvField.server_start_time.name: "argrelay.check_env server_start_time",
                },
            },
        )

    def verify_online_value(
        self,
        field_name,
        field_value,
    ) -> CheckEnvResult:
        check_env_result: CheckEnvResult = super().verify_online_value(
            field_name,
            field_value,
        )
        # Put formated time to the message:
        if check_env_result.result_category.VerificationSuccess:
            utc_time = datetime.fromtimestamp(int(field_value), timezone.utc)
            epoch_start = datetime.fromtimestamp(0, timezone.utc)

            # Local timezone:
            abs_time_str = utc_time.astimezone(None).isoformat()
            rel_time_str = format_time_to_relative(
                time.time() * 1000,
                (utc_time - epoch_start).total_seconds() * 1000,
            )

            check_env_result.result_message = f"{abs_time_str} ~ {rel_time_str}"
        return check_env_result



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)
from typing import Union

from argrelay_api_server_cli.schema_response.AssignedValue import AssignedValue
from argrelay_app_server.runtime_context.SearchControl import SearchControl
from argrelay_lib_root.enum_desc.ValueSource import ValueSource


@dataclass
class EnvelopeContainer:
    """
    Combination of relevant config, user input, and the last `data_envelope` found based on that.

    See also :class:`EnvelopeContainerSchema`.
    """

    search_control: SearchControl = field(default_factory = lambda: SearchControl())
    """
    A specs how to query `data_envelope` for this `EnvelopeContainer`.
    """

    data_envelopes: list[dict] = field(default_factory = lambda: [])
    """
    These `data_envelope`-s are based on `search_control` for this `EnvelopeContainer` (and command line input).

    It contains only 0 or 1 `data_envelope` for performance reasons in case of:
    *   `ServerAction.ProposeArgValues` (Tab-completion)
    *   `ServerAction.DescribeLineArgs` (Alt+Shift+Q)
    See `QueryEngine.query_prop_values` and `QueryResult.data_envelopes`.

    It contains all N `data_envelope`-s in case of `ServerAction.RelayLineArgs`.
    See `QueryResult.data_envelopes`.
    """

    found_count: int = field(default = 0)
    """
    Counter of `data_envelope`-s found in the last query.
    """

    used_token_bucket: Union[int, None] = field(default = None)
    """
    FS_97_64_39_94 `token_bucket` (index) where args were consumed from for the current `envelope_container`.
    """

    # TODO: TODO_55_51_89_92: review and update `args_context`
    # TODO: Part of (or actually is?) `args_context`: FS_62_25_92_06:
    assigned_prop_name_to_prop_value: dict[str, AssignedValue] = field(default_factory = lambda: {})
    """
    All assigned args (from interpreted tokens) mapped as type:value which belong to `data_envelope`.
    """

    # TODO: TODO_55_51_89_92: review and update `args_context`
    # TODO: Part of (or not?) `args_context` (FS_62_25_92_06) to support FS_13_51_07_97 (single out implicit values):
    remaining_prop_name_to_prop_value: dict[str, list[str]] = field(default_factory = lambda: {})
    """
    All `prop_value`-s per `prop_name` left as options to match `arg_value` given on the command line.

    When `arg_value` from command line matches one of the `prop_value`-s for one of the `prop_name`,
    this `prop_name` moves to `assigned_prop_name_to_prop_value`.
    """

    filled_prop_values_hidden_by_defaults: dict[str, list[str]] = field(default_factory = lambda: {})
    """
    FS_72_53_55_13: values hidden by defaults:
    This dict stores values which were hidden by applying defaults (see FS_72_40_53_00 fill control).
    Only keys for those `prop_name`-s which have `ValueSource.default_value` in
    `assigned_prop_name_to_prop_value` are listed here.
    """

    def populate_implicit_arg_values(
        self,
    ) -> bool:
        """
        When possible `arg_value`-s are singled out, assign them as `ValueSource.implicit_value`.

        When `data_envelope` is singled out, all remaining single-value `prop_name`-s become `ValueSource.implicit_value`.

        Implements FS_13_51_07_97 singled out implicit values.

        Return:
        *   True if any value was assigned.
        *   False otherwise.
        """

        any_assigned = False
        # Filter as in: FS_31_70_49_15 search control:
        for prop_name in self.search_control.arg_name_to_prop_name_dict.values():
            if prop_name not in self.assigned_prop_name_to_prop_value:
                if prop_name in self.remaining_prop_name_to_prop_value:
                    prop_values = self.remaining_prop_name_to_prop_value[prop_name]

                    # FS_06_99_43_60 array `prop_value`:
                    # When a `data_envelope` is singled out, its array `prop_value` will not be set as
                    # singled out via `ValueSource.implicit_value` (user can still select them explicitly).
                    # This is deliberate for now as the selection of specific value out of the array can be used
                    # by delegators to implement difference in behavior.

                    if len(prop_values) == 1:
                        del self.remaining_prop_name_to_prop_value[prop_name]
                        self.assigned_prop_name_to_prop_value[prop_name] = AssignedValue(
                            # single value:
                            prop_values[0],
                            ValueSource.implicit_value,
                        )
                        any_assigned = True
        return any_assigned



import os
import time
import typing
from random import randrange

from argrelay_app_client.client_spec.ShellContext import ShellContext
from argrelay_app_client.relay_client.ClientRemote import ClientRemote
from argrelay_app_client.relay_client.proc_splitter import has_child_exited
from argrelay_lib_root.enum_desc.ProcRole import ProcRole
from argrelay_lib_root.enum_desc.TermColor import TermColor
from argrelay_lib_root.misc_helper_common import eprint

# Use regular Tab indent size:
spinner_length: int = 4

# Rotate spinner state every:
spinner_frame_sec: float = 0.1

# Results of multiple stream reads:
child_data_chunks: list = []

child_data_chunk_max_size: int = 1024


def spin_wait_for_child(
    r_pipe_end: typing.BinaryIO,
    shell_ctx: ShellContext,
    spinless_sleep_sec: float,
):
    """
    Display spinner while child request is running.
    """

    # Ensure we can read from the pipe without blocking if there is no input ready:
    os.set_blocking(r_pipe_end.fileno(), False)

    # TODO: Try busy spin for the `spinless_sleep_sec` instead of sleeping?
    # TODO: Maybe, once spinner started, spin even while pumping?

    # Prevent spinner on instant replies - have an initial sleep first
    # before drawing spinner for the first time
    # (this sleep will exit prematurely if the child exits):
    if not has_child_exited():
        # There is a chance that child exits here (after the check above but before the sleep below).
        # Ignore that - it is unlikely and, when it happens, the worst case is a needless sleep.
        time.sleep(spinless_sleep_sec)
    if has_child_exited():
        # Pump out the rest of data without spinner:
        return

    # We waited enough - start spinner:
    pending_cursor = generate_pending_cursor()
    is_read_successful: bool = read_next_child_data_chunk(r_pipe_end)
    while (
        # Ensure first that no input is ready.
        # Child block on pipe write (and never exits) if parent does not read it:
        is_read_successful
        or
        not has_child_exited()
    ):
        # Print spinner state chars:
        eprint(next(pending_cursor), end = "", flush = True)
        if not is_read_successful:
            # Sleep only while read failed -
            # if child is still running and read is successful,
            # keep pumping data out of the pipe without sleeping:
            time.sleep(spinner_frame_sec)
        # Step backward:
        eprint("\b" * spinner_length, end = "")
        is_read_successful: bool = read_next_child_data_chunk(r_pipe_end)
        if has_child_exited():
            # Pump out the rest of data without spinner:
            break

    # Overwrite spinner state chars by expected content from the prev command line
    # (padded with spaces, if beyond the end of the line):
    first_cpos = shell_ctx.cursor_cpos
    last_cpos = shell_ctx.cursor_cpos + spinner_length
    line_portion = shell_ctx.command_line[first_cpos:last_cpos]
    pad_length = spinner_length - len(line_portion)
    eprint(line_portion + " " * pad_length, end = "")
    # Step backward:
    eprint("\b" * spinner_length, end = "", flush = True)


def generate_pending_cursor():
    cursor_states = [
        f"{TermColor.spinner_color.value}/|\\|{TermColor.reset_style.value}",
        f"{TermColor.spinner_color.value}|/|\\{TermColor.reset_style.value}",
        f"{TermColor.spinner_color.value}\\|/|{TermColor.reset_style.value}",
        f"{TermColor.spinner_color.value}|\\|/{TermColor.reset_style.value}",
    ]
    # Use random start state:
    random_shift = randrange(len(cursor_states))
    shifted_states = cursor_states[random_shift:] + cursor_states[:random_shift]
    while True:
        for cursor_state in shifted_states:
            yield cursor_state


def read_next_child_data_chunk(
    r_pipe_end: typing.BinaryIO,
) -> bool:
    child_data_chunk = r_pipe_end.read(child_data_chunk_max_size)
    if child_data_chunk:
        child_data_chunks.append(child_data_chunk)
        return True
    else:
        return False


def spinner_main(
    call_ctx,
    client_config,
    proc_role: ProcRole,
    is_optimized_completion,
    r_pipe_end,
):
    spin_wait_for_child(
        r_pipe_end,
        call_ctx,
        client_config.spinless_sleep_sec,
    )

    while read_next_child_data_chunk(r_pipe_end):
        pass

    abstract_client = ClientRemote(
        client_config,
        proc_role,
        None,
        is_optimized_completion,
        -1,
    )
    # TODO: Rethink/Rename: spinner does not make any request:
    #       Maybe remove *Client altogether and move its functionality into *ClientCommandFactory?
    #       Then simply execute command here?
    abstract_client.make_request(call_ctx)



from typing import Union

from argrelay_lib_root.enum_desc.CompType import CompType
from argrelay_test_infra.test_infra import parse_line_and_cpos
from argrelay_test_infra.test_infra.TestCase import TestCase


class ShellInputTestCase(TestCase):

    def __init__(
        self,
        line_no: int,
        case_comment: str,
    ):
        super().__init__(
            line_no,
            case_comment,
        )
        self.test_line: Union[str, None] = None
        self.command_line: Union[str, None] = None
        self.cursor_cpos: Union[int, None] = None
        self.comp_type: Union[CompType, None] = None

    def set_command_line(
        self,
        given_command_line: str,
    ):
        assert self.command_line is None
        self.command_line = given_command_line

    def set_cursor_cpos(
        self,
        given_cursor_cpos: int,
    ):
        assert self.cursor_cpos is None
        self.cursor_cpos = given_cursor_cpos

    def set_test_line(
        self,
        given_test_line: str,
    ):
        assert self.test_line is None
        assert self.command_line is None
        assert self.cursor_cpos is None
        self.test_line = given_test_line
        (
            self.command_line,
            self.cursor_cpos,
        ) = parse_line_and_cpos(given_test_line)

    def set_comp_type(
        self,
        given_comp_type: CompType,
    ):
        assert self.comp_type is None
        self.comp_type = given_comp_type



from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.InterpStep import InterpStep
from argrelay_lib_server_plugin_core.plugin_interp.AbstractInterpFactory import (
    AbstractInterp,
    AbstractInterpFactory,
)
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig


class NoopInterpFactory(AbstractInterpFactory):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> "NoopInterp":
        # `NoopInterpFactory` is not normally attached to any tree and
        # `load_func_envelopes` is not invoked to clone/populate separate configs
        # (which makes it use `plugin_config_dict` directly instead of `interp_tree_node_config_dict`) -
        # use `plugin_config_dict` directly if so:
        if interp_ctx.interp_tree_abs_path not in self.interp_tree_abs_paths_to_node_configs:
            plugin_config_dict = self.plugin_config_dict
        else:
            plugin_config_dict = self.interp_tree_abs_paths_to_node_configs[interp_ctx.interp_tree_abs_path]
        return NoopInterp(
            self.plugin_instance_id,
            plugin_config_dict,
            interp_ctx,
        )


class NoopInterp(AbstractInterp):
    """
    Interpreter which does nothing

    It stops processing of the command line on itself (as `next_interp` returns `None` by default)
    """

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp



from __future__ import annotations

from copy import deepcopy

from marshmallow import (
    fields,
    post_load,
    RAISE,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.runtime_data_server_plugin.PluginConfig import PluginConfig
from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry
from argrelay_schema_config_server.schema_config_server_plugin.BasePluginConfigSchema import (
    BasePluginConfigSchema,
    serialize_dag_to_list,
)
from argrelay_schema_config_server.schema_config_server_plugin.PluginEntrySchema import plugin_entry_desc

server_plugin_instances_ = "server_plugin_instances"
server_plugin_instance_groups_ = "server_plugin_instance_groups"


class PluginConfigSchema(BasePluginConfigSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = PluginConfig

    # TODO: consider removing this and using only `server_plugin_instance_groups`:
    # Plugin config data: key = `plugin_instance_id`, value = `plugin_entry`:
    server_plugin_instances = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(plugin_entry_desc.dict_schema),
        required = False,
        load_default = {},
    )

    # TODO: consider removing `server_plugin_instances` and using this:
    # Same as `server_plugin_instances` but with extra `dict` level
    # to group plugin instances under that level.
    # This has no impact on the logic - it is purely a convenience for configuration
    # (e.g. to be able to `!include` entire group from another file - see FS_70_55_40_99 splitting config file).
    server_plugin_instance_groups = fields.Dict(
        keys = fields.String(),
        values = fields.Dict(
            keys = fields.String(),
            values = fields.Nested(plugin_entry_desc.dict_schema),
            required = True,
        ),
        required = False,
        load_default = {},
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        input_dict = self.adjust_for_unification(input_dict)

        # TODO: avoid calling it on load - it is only needed by server, not by client:
        # Build DAG:
        plugin_instance_id_activate_order_dag = {}
        for plugin_instance_id in input_dict[server_plugin_instances_]:
            plugin_entry: PluginEntry = input_dict[server_plugin_instances_][plugin_instance_id]
            plugin_instance_id_activate_order_dag[plugin_instance_id] = plugin_entry.plugin_dependencies

        return type(self).model_class(
            **input_dict,
            plugin_instance_id_activate_list = serialize_dag_to_list(
                plugin_instance_id_activate_order_dag,
            ),
        )

    # noinspection PyMethodMayBeStatic
    def adjust_for_unification(
        self,
        input_dict,
    ):
        # Make a `deepcopy` as `marshmallow` reuses instances specified in `load_default`:
        input_dict = deepcopy(input_dict)

        # Move all plugin instances from `server_plugin_instance_groups` to `server_plugin_instances`
        server_plugin_instance_groups = input_dict.pop(server_plugin_instance_groups_)
        dst_dict: dict = input_dict[server_plugin_instances_]
        src_dict: dict
        for src_dict in server_plugin_instance_groups.values():
            for plugin_instance_id, plugin_entry in src_dict.items():
                if plugin_instance_id in dst_dict:
                    raise AssertionError(
                        f"`plugin_instance_id` [{plugin_instance_id}] "
                        f"is already defined in `{server_plugin_instances_}` "
                    )
                dst_dict[plugin_instance_id] = plugin_entry
            src_dict.clear()
        return input_dict


plugin_config_desc = TypeDesc(
    dict_schema = PluginConfigSchema(),
    ref_name = PluginConfigSchema.__name__,
    dict_example = {
        server_plugin_instances_: {
            "some_plugin_instance_id": plugin_entry_desc.dict_example,
            "some_plugin_instance_id1": plugin_entry_desc.dict_example,
            "some_plugin_instance_id2": plugin_entry_desc.dict_example,
        },
    },
    default_file_path = "argrelay_plugin.yaml",
)



import inspect
from typing import Union

from argrelay_api_plugin_server_abstract.ConfiguratorAbstract import ConfiguratorAbstract
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig


class ConfiguratorConsistent(ConfiguratorAbstract):
    """
    Query all values on start (on activation) and consistently reply the same value on next queries.

    This is to ensure that, for example, the same git commit id is reported
    matching server runtime rather than current disk state.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.consistent_replies: dict[str, Union[str, None]] = {}

    def activate_plugin(
        self,
    ) -> None:
        """
        Call all funcs to store their values for subsequent calls.
        """
        for func_obj in [
            self.provide_project_title,
            self.provide_project_page_url,
            self.provide_project_git_commit_display_string,
            self.provide_project_git_files_by_commit_id_url_prefix,
            self.provide_project_commit_id_url_prefix,
            self.provide_project_git_commit_id,
            self.provide_project_git_commit_time,
            self.provide_project_git_repo_relative_argrelay_dir,
            self.provide_project_current_config_path,
        ]:
            func_obj()

    def reply_consistently(
        self,
    ):
        """
        If replied once with one answer, repeat this answer subsequently.
        """

        caller_func_name = inspect.currentframe().f_back.f_code.co_name
        if caller_func_name not in self.consistent_replies:
            # Call "protected" func to get value:
            func_reply = getattr(self, f"_{caller_func_name}")()
            self.consistent_replies[caller_func_name] = func_reply
            return func_reply
        else:
            return self.consistent_replies[caller_func_name]

    ####################################################################################################################

    def provide_project_title(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_title(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_page_url(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_page_url(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_git_commit_display_string(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_git_commit_display_string(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_git_files_by_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_git_files_by_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_git_commit_id(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_git_commit_id(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_git_commit_time(
        self,
    ) -> Union[int, None]:
        return self.reply_consistently()

    def _provide_project_git_commit_time(
        self,
    ) -> Union[int, None]:
        return None

    ####################################################################################################################

    def provide_project_git_repo_relative_argrelay_dir(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_git_repo_relative_argrelay_dir(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################

    def provide_project_current_config_path(
        self,
    ) -> Union[str, None]:
        return self.reply_consistently()

    def _provide_project_current_config_path(
        self,
    ) -> Union[str, None]:
        return None

    ####################################################################################################################



"""
"""
from __future__ import annotations

import os
import subprocess
from typing import Union


def get_git_repo_root_path(
    git_repo_path,
) -> Union[str, None]:
    if not os.path.isdir(git_repo_path):
        # A git repo can only be inside a dir:
        return None

    sub_proc = subprocess.run(
        [
            "git",
            "-C",
            git_repo_path,
            "rev-parse",
            "--show-toplevel",
        ],
        cwd = git_repo_path,
        capture_output = True,
    )
    exit_code = sub_proc.returncode

    if exit_code == 0:
        return sub_proc.stdout.decode("utf-8").strip()
    else:
        return None


def is_git_repo(
    git_repo_path,
) -> bool:
    """
    Detects whether `git_repo_path` points within a git repository.

    It is not required for `git_repo_path` to match git repo root (any sub-dir is accepted).
    """
    git_repo_root_path = get_git_repo_root_path(git_repo_path)
    if git_repo_root_path is None:
        return False
    else:
        return True


def get_full_commit_id(
    git_repo_path: str,
) -> str:
    sub_proc = subprocess.run(
        [
            "git",
            "rev-parse",
            "HEAD",
        ],
        cwd = git_repo_path,
        capture_output = True,
    )
    exit_code = sub_proc.returncode

    if exit_code == 0:
        return sub_proc.stdout.decode("utf-8").strip()
    else:
        # If `is_git_repo` returns `True`, this should not happen:
        raise RuntimeError


def get_short_commit_id(
    git_repo_path: str,
) -> str:
    sub_proc = subprocess.run(
        [
            "git",
            "rev-parse",
            "--short",
            "HEAD",
        ],
        cwd = git_repo_path,
        capture_output = True,
    )
    exit_code = sub_proc.returncode

    if exit_code == 0:
        return sub_proc.stdout.decode("utf-8").strip()
    else:
        # If `is_git_repo` returns `True`, this should not happen:
        raise RuntimeError


def get_commit_time(
    git_repo_path: str,
) -> int:
    """
    Returns seconds since epoch (Unix time) for the last commit.
    """

    sub_proc = subprocess.run(
        [
            "git",
            "log",
            "-1",
            "--format=%ct",
        ],
        cwd = git_repo_path,
        capture_output = True,
    )
    exit_code = sub_proc.returncode

    if exit_code == 0:
        return int(sub_proc.stdout.decode("utf-8").strip())
    else:
        # If `is_git_repo` returns `True`, this should not happen:
        raise RuntimeError


def get_unversioned_file_list(
    base_dir: str,
) -> list[str]:
    sub_proc = subprocess.run(
        [
            "git",
            "ls-files",
            "--others",
            "--exclude-standard",
        ],
        cwd = base_dir,
        capture_output = True,
    )
    exit_code = sub_proc.returncode

    if exit_code == 0:
        return sub_proc.stdout.decode("utf-8").splitlines()
    else:
        # If `is_git_repo` returns `True`, this should not happen:
        raise RuntimeError



from marshmallow import (
    fields,
    RAISE,
    Schema,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_server_plugin_core.plugin_interp.InterpTreeInterpFactoryConfigSchema import ignored_func_ids_list_


class FirstArgInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    # TODO: FS_42_76_93_51: See todo on the same field in `InterpTreeInterpFactoryConfigSchema`:
    ignored_func_ids_list = fields.List(
        fields.String(),
        required = False,
        load_default = [],
    )


first_arg_interp_factory_config_example = {
    ignored_func_ids_list_: [
        "func_id_some_ignored",
    ],
}
first_arg_interp_factory_config_desc = TypeDesc(
    dict_schema = FirstArgInterpFactoryConfigSchema(),
    ref_name = FirstArgInterpFactoryConfigSchema.__name__,
    dict_example = first_arg_interp_factory_config_example,
    default_file_path = "",
)



import json
import socket

from argrelay_app_client.client_command_remote.ClientCommandRemoteWorkerAbstract import (
    ClientCommandRemoteWorkerAbstract,
)
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay_lib_root.schema_config.ConnectionConfig import ConnectionConfig


class ClientCommandRemoteWorkerTextProposeArgValuesOptimized(ClientCommandRemoteWorkerAbstract):
    """
    This class optimizes what can otherwise be done via generic `ClientCommandRemoteWorkerJson`.

    See `completion_perf_notes.md`.

    Importing many packages including `marshmallow` slows down startup and responses on `Tab` requests.

    Performance is not critical for other client commands
    (e.g. `ServerAction.DescribeLineArgs` or `ServerAction.RelayLineArgs`),
    but it is critical for `Tab` (`ServerAction.ProposeArgValues`).

    Because `Tab`-completion is latency-sensitive, `ServerAction.ProposeArgValues` uses this specialized implementation.
    The drawback is that it also requires special maintenance/testing.
    """

    # TODO: TODO_72_51_13_18: test optimized requests:
    #       Write mocked test to cover internal logic like function `recvall` below.

    def _execute_remote_call(
        self,
    ):

        with socket.socket(
            socket.AF_INET,
            socket.SOCK_STREAM,
        ) as s:
            self.execute_command_with_socket(
                s,
                self.curr_connection_config,
            )

    def execute_command_with_socket(
        self,
        s,
        connection_config: ConnectionConfig,
    ):
        s.connect((
            connection_config.server_host_name,
            connection_config.server_port_number,
        ))

        request_body_str = (f"""\
{{
    "client_version": {json.dumps(self.call_ctx.client_version)},
    "client_conf_target": {json.dumps(self.call_ctx.client_conf_target)},
    "server_action": "{self.call_ctx.server_action.name}",
    "command_line": {json.dumps(self.call_ctx.command_line)},
    "cursor_cpos": {self.call_ctx.cursor_cpos},
    "comp_scope": "{self.call_ctx.comp_scope.name}",
    "client_uid": {json.dumps(self.call_ctx.client_uid)},
    "client_pid": {self.call_ctx.client_pid},
    "is_debug_enabled": {'true' if self.call_ctx.is_debug_enabled else 'false'}
}}
""")
        request_body_len = len(request_body_str.encode())

        request_str = (f"""\
POST {ServerAction.ProposeArgValues.value} HTTP/1.1\r
Content-Type: application/json\r
Content-Length: {request_body_len}\r
Connection: close\r
\r
{request_body_str}
""")
        ElapsedTime.measure("before_request")

        s.sendall(request_str.encode())
        response_str = self.recvall(s).decode()

        ElapsedTime.measure("after_request")

        # First line, second space-delimited substring:
        # HTTP/1.1 200 OK\r\n
        response_status_line = response_str[:response_str.find("\r")]
        first_space_cpos = response_status_line.find(" ")
        response_status_code = int(
            response_status_line[first_space_cpos + 1:first_space_cpos + 1 + 3]
        )
        # Content after headers (after empty line):
        content_cpos = response_str.find("\r\n\r\n") + 4
        if content_cpos < len(response_str):
            response_body_str = response_str[content_cpos:]
        else:
            response_body_str = ""

        ElapsedTime.measure("after_deserialization")

        try:
            if response_status_code == 200:
                self.bytes_src.accept_bytes(response_body_str.encode())
            else:
                self.raise_error(response_status_code)
        finally:
            ElapsedTime.measure("after_handle_response")

    # noinspection SpellCheckingInspection
    @staticmethod
    def recvall(s):
        bytes_parts = []
        while True:
            bytes_part = s.recv(1000)
            if not bytes_part:
                break
            bytes_parts.append(bytes_part)
        return b"".join(bytes_parts)



from marshmallow import (
    fields,
    RAISE,
)

from argrelay_api_server_cli.schema_response.AssignedValue import AssignedValue
from argrelay_lib_root.enum_desc.ValueSource import ValueSource
from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc

prop_value_ = "prop_value"
value_source_ = "value_source"


class AssignedValueSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = AssignedValue

    prop_value = fields.String(
        required = True,
    )

    value_source = fields.Enum(
        ValueSource,
        by_value = False,
        required = True,
    )


assigned_value_desc = TypeDesc(
    dict_schema = AssignedValueSchema(),
    ref_name = AssignedValueSchema.__name__,
    dict_example = {
        prop_value_: "value_1",
        value_source_: ValueSource.implicit_value.name,
    },
    default_file_path = "",
)



"""
This module lists concreate implementations of `AbstractArg`-s.
"""

from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_app_server.runtime_context.AbstractArg import (
    ArgCommand,
    ArgCommandIncomplete,
    ArgCommandValueDictated,
    ArgCommandValueOffered,
)


@dataclass
class ArgCommandData(ArgCommand):
    """
    Implements `command_arg` - see FS_27_16_67_19 line syntax
    """

    token_ipos_list: list[int] = field()
    """
    List of all token ipos-es (indexes) belonging to that arg.
    
    These ipos-es reference entries within `ParsedContext.all_tokens`.
    """

    def get_arg_tokens(
        self,
    ) -> list[int]:
        return self.token_ipos_list


@dataclass
class ArgCommandDataValue(ArgCommandData):
    """
    Extends `ArgCommandData` with `arg_value`.
    """

    arg_value: str = field()

    def get_arg_value(
        self
    ) -> str:
        return self.arg_value


@dataclass
class ArgCommandDataValueOffered(ArgCommandDataValue, ArgCommandValueOffered):
    """
    Implements FS_96_46_42_30 `offered_arg`.
    """


@dataclass
class ArgCommandDataValueDictated(ArgCommandDataValue, ArgCommandValueDictated):
    """
    Implements FS_20_88_05_60 `dictated_arg``.
    """

    arg_name: str = field()

    def get_arg_name(
        self
    ) -> str:
        return self.arg_name


@dataclass
class ArgCommandDataIncomplete(ArgCommandData, ArgCommandIncomplete):
    """
    Implements FS_08_58_30_24 `incomplete_arg`.
    """

    arg_name: str = field()

    def get_arg_name(
        self
    ) -> str:
        return self.arg_name



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

index_props_ = "index_props"


@dataclass(frozen = True)
class IndexModel:
    """
    Defines index model for FS_45_08_22_15 index model API.
    """

    collection_name: str

    index_props: list[str] = field()
    """
    Lists fields of `data_envelope` which are to be indexed by MongoDB for given collection.
    """



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)


@dataclass(frozen = True)
class EnvelopeCollection:
    """
    Allows distributing `data_envelope`-s across FS_56_43_05_79 different collections on data load.

    See also `EnvelopeCollectionSchema`.
    """

    collection_name: str = field()

    data_envelopes: list = field()



from argrelay_app_client.client_pipeline.BytesSrcAbstract import BytesSrcAbstract


class BytesSrcSender(BytesSrcAbstract):

    def __init__(
        self,
        w_pipe_end: "typing.BinaryIO",
    ):
        super().__init__(
        )
        self.w_pipe_end: "typing.BinaryIO" = w_pipe_end

    def accept_bytes(
        self,
        given_bytes: bytes,
    ):
        self.w_pipe_end.write(given_bytes)



from marshmallow import (
    fields,
    INCLUDE,
    validates_schema,
)

from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    DataEnvelopeSchema,
    envelope_id_,
    envelope_payload_,
    instance_data_,
    sample_prop_name_a_,
    sample_prop_name_b_,
    sample_prop_name_c_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    function_envelope_instance_data_desc,
)


class FuncEnvelopeSchema(DataEnvelopeSchema):
    class Meta:
        # Same as in base `DataEnvelopeSchema` - all unknown fields used as search props:
        unknown = INCLUDE
        strict = True

    envelope_payload = fields.Dict(
        required = False,
    )

    @validates_schema
    def validate_known(
        self,
        input_dict: dict,
        **kwargs,
    ):
        assert input_dict.get(ReservedPropName.envelope_class.name, None) == ReservedEnvelopeClass.class_function.name
        function_envelope_instance_data_desc.validate_dict(input_dict[instance_data_])


func_id_some_func_ = "func_id_some_func"

func_envelope_desc = TypeDesc(
    dict_schema = DataEnvelopeSchema(),
    ref_name = DataEnvelopeSchema.__name__,
    dict_example = {
        # TODO: Do we need to duplicate `envelope_id_` and `ReservedPropName.func_id.name`?
        envelope_id_: func_id_some_func_,
        instance_data_: function_envelope_instance_data_desc.dict_example,
        envelope_payload_: {},
        sample_prop_name_a_: "sample_prop_value_1",
        sample_prop_name_b_: "sample_prop_value_2",
        sample_prop_name_c_: "sample_prop_value_3",
        ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
        ReservedPropName.help_hint.name: f"Some help hint",
        ReservedPropName.func_state.name: FuncState.fs_demo.name,
        ReservedPropName.func_id.name: func_id_some_func_,
    },
    default_file_path = "",
)



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)


@dataclass
class InitControl:
    """
    Implements `init_control` (FS_46_96_59_05).
    """

    init_types_to_values: dict[str, str] = field()



from __future__ import annotations

from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialFunc import SpecialFunc
from argrelay_lib_server_plugin_core.enum_desc.NoDataPropName import NoDataPropName
from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorNoopBase import DelegatorNoopBase
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
)
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import populate_search_control
from argrelay_schema_config_server.schema_config_server_app.CompositeNodeSchema import func_id_


class DelegatorNoopNoDataFunc(DelegatorNoopBase):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        no_data_search_control = populate_search_control(
            ReservedEnvelopeClass.class_no_data.name,
            {
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_no_data.name,
            },
            [
                # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                {"class": ReservedPropName.envelope_class.name},

                {"no_data_arg_name_one": NoDataPropName.no_data_prop_name_one.name},
                {"no_data_arg_name_two": NoDataPropName.no_data_prop_name_two.name},
            ],
        )

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: SpecialFunc.func_id_no_data.name,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        no_data_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
                ReservedPropName.help_hint.name: "Function which has no data it searches for and does nothing",
                ReservedPropName.func_state.name: FuncState.fs_ignorable.name,
                ReservedPropName.func_id.name: SpecialFunc.func_id_no_data.name,
            },
        ]
        return func_envelopes



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_api_server_cli.schema_response.InterpResult import InterpResult
from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry


@dataclass
class InvocationInput(InterpResult):
    """
    See also :class:`InvocationInputSchema`.

    Unlike `InterpResult` this class provides full list of `data_envelope`-s per `envelope_container`
    because invocation is not latency-sensitive and funcs expect all `data_envelope`-s.
    """

    call_ctx: CallContext = field()
    """
    Call context (sent back to client) which caused this `invocation_input`.
    """

    delegator_plugin_entry: PluginEntry = field()
    """
    The `PluginEntry` taken from config on the server side to let client invoke that plugin.

    It is assumed that server and client code, if not of the same version, at least, compatible.
    """

    custom_plugin_data: dict = field()
    """
    A placeholder dict for exclusive use by plugin.

    Whatever plugin server side needs to tell its plugin peer on the client side.
    """

    @staticmethod
    def with_interp_context(
        interp_ctx: InterpContext,
        delegator_plugin_entry: PluginEntry,
        custom_plugin_data: dict,
    ) -> InvocationInput:
        return InvocationInput(
            call_ctx = interp_ctx.parsed_ctx,
            arg_values = interp_ctx.comp_suggestions,
            all_tokens = interp_ctx.parsed_ctx.all_tokens,
            excluded_tokens = interp_ctx.excluded_tokens,
            consumed_token_buckets = interp_ctx.consumed_token_buckets,
            envelope_containers = interp_ctx.envelope_containers,
            tan_token_ipos = interp_ctx.parsed_ctx.tan_token_ipos,
            tan_token_l_part = interp_ctx.parsed_ctx.tan_token_l_part,
            delegator_plugin_entry = delegator_plugin_entry,
            custom_plugin_data = custom_plugin_data,
        )



from __future__ import annotations

from copy import deepcopy

from argrelay_api_plugin_server_abstract.AbstractLoader import AbstractLoader
from argrelay_app_server.relay_server.IndexModel import IndexModel
from argrelay_app_server.relay_server.QueryEngine import QueryEngine
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_core.plugin_loader.ConfigOnlyLoaderConfigSchema import (
    collection_name_to_index_props_map_,
    config_only_loader_config_desc,
    data_envelopes_,
    envelope_class_to_collection_name_map_,
)
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig

_primitive_types = (bool, str, int, float)


def convert_envelope_fields_to_string(
    data_envelope: dict,
):
    """
    Converts primitive types to `str`.

    See also:
    https://stackoverflow.com/a/6392016/441652
    """
    for prop_name, prop_value in data_envelope.items():
        if isinstance(prop_value, _primitive_types):
            data_envelope[prop_name] = str(prop_value)
    return data_envelope


class ConfigOnlyLoader(AbstractLoader):
    """
    Implements FS_49_96_50_77 config_only_loader.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.plugin_config_dict = config_only_loader_config_desc.obj_from_input_dict(self.plugin_config_dict)

    def list_index_models(
        self,
    ) -> list[IndexModel]:

        collection_name_to_index_props_map: dict[str, list[str]] = self.plugin_config_dict[
            collection_name_to_index_props_map_
        ]
        index_models: list[IndexModel] = []
        for collection_name, index_props in collection_name_to_index_props_map.items():
            index_models.append(IndexModel(
                collection_name = collection_name,
                index_props = index_props,
            ))

        return index_models

    def load_envelope_collections(
        self,
        query_engine: QueryEngine,
    ) -> list[EnvelopeCollection]:

        # FS_49_96_50_77 config_only_loader plugin:
        # *   `envelope_class_to_collection_name_map`:
        envelope_class_to_collection_name_map: dict[str, str] = self.plugin_config_dict[
            envelope_class_to_collection_name_map_
        ]
        # *   list of `data_envelope`-s (actual data):
        data_envelopes = [
            convert_envelope_fields_to_string(data_envelope)
            for data_envelope in deepcopy(self.plugin_config_dict[data_envelopes_])
        ]

        envelope_collections: dict[str, EnvelopeCollection] = {}

        for data_envelope in data_envelopes:
            class_name = data_envelope.get(
                ReservedPropName.envelope_class.name,
                ReservedEnvelopeClass.ClassUnknown.name,
            )
            collection_name = envelope_class_to_collection_name_map.get(class_name, class_name)

            envelope_collection = envelope_collections.setdefault(
                collection_name,
                EnvelopeCollection(
                    collection_name = collection_name,
                    data_envelopes = [],
                ),
            )

            envelope_collection.data_envelopes.append(data_envelope)

        return envelope_collections.values()



from argrelay_app_server.relay_server.QueryEngine import QueryEngine
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName


class HelpHintCache:
    """
    Implements FS_94_30_49_28 help hint.
    """

    def __init__(
        self,
        query_engine: QueryEngine,
    ):
        self.query_engine: QueryEngine = query_engine
        self.help_hint_dict: dict = {}

    def populate_cache(self):

        help_hint_envelopes = self.query_engine.query_data_envelopes(
            ReservedEnvelopeClass.class_help.name,
            {
                f"{ReservedPropName.envelope_class.name}": f"{ReservedEnvelopeClass.class_help.name}",
            },
        )

        for help_hint_envelope in help_hint_envelopes:
            prop_name = help_hint_envelope[ReservedPropName.prop_name.name]
            prop_value = help_hint_envelope[ReservedPropName.prop_value.name]
            help_hint = help_hint_envelope[ReservedPropName.help_hint.name]
            if prop_name not in self.help_hint_dict:
                self.help_hint_dict[prop_name] = {}
            self.help_hint_dict[prop_name][prop_value] = help_hint

    def get_value_with_help_hint(
        self,
        prop_name: str,
        prop_value: str,
    ) -> str:
        if prop_name in self.help_hint_dict:
            if prop_value in self.help_hint_dict[prop_name]:
                return f"{prop_value} # {self.help_hint_dict[prop_name][prop_value]}"
        return prop_value



from __future__ import annotations

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract


class DelegatorCheckEnvBase(DelegatorSingleFuncAbstract):
    """
    Implements base functionality used by FS_36_17_84_44 `check_env`.
    """



from __future__ import annotations

from copy import deepcopy

from marshmallow import (
    fields,
    pre_load,
    RAISE,
    Schema,
    validates_schema,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_server_plugin_core.plugin_delegator.FuncConfigSchema import (
    func_config_desc,
    func_envelope_,
)
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaConfigDelegatorConfigBase import (
    func_configs_,
    SchemaConfigDelegatorConfigBase,
)
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import envelope_payload_
from argrelay_schema_config_server.schema_config_interp.FuncEnvelopeSchema import (
    func_id_some_func_,
)

command_template_ = "command_template"
echo_command_on_stderr_ = "echo_command_on_stderr"


class SchemaConfigDelegatorConfigOnly(SchemaConfigDelegatorConfigBase):
    """
    Part of FS_49_96_50_77 config_only_delegator implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    @pre_load
    def pre_load(
        self,
        input_dict: dict,
        **kwargs
    ) -> dict:
        """
        See `validate_data` why this has to be done:
        """
        for func_config in input_dict[func_configs_].values():
            func_config[func_envelope_][
                envelope_payload_
            ] = config_only_delegator_envelope_payload_desc.dict_from_input_dict(
                func_config[func_envelope_][envelope_payload_]
            )
        return input_dict

    @validates_schema
    def validate_data(
        self,
        input_dict: dict,
        **kwargs,
    ):
        """
        `SchemaConfigDelegatorConfigBase.func_configs` dict contains values with
        `FuncConfigSchema.func_envelope` of `FuncEnvelopeSchema` which allows arbitrary dict in `envelope_payload`.

        `SchemaConfigDelegatorConfigOnly` expects them to be of `DelegatorConfigOnlyEnvelopePayloadSchema`.
        """
        for func_config in input_dict[func_configs_].values():
            config_only_delegator_envelope_payload_desc.validate_dict(func_config[func_envelope_][envelope_payload_])


class DelegatorConfigOnlyEnvelopePayloadSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    command_template = fields.String(
        required = True,
    )

    echo_command_on_stderr = fields.Boolean(
        required = False,
        load_default = True,
    )

    # TODO: TODO_54_68_18_12: Support defaults for config-only delegator:
    #       Add a schema for this config but likely into _new_ `DelegatorConfigBaseEnvelopePayloadSchema`
    #       as it can be used with other configurable plugins even if they do not use
    #       `command_template` (as this `DelegatorConfigOnlyEnvelopePayloadSchema` class does).

    # TODO: TODO_74_73_60_93: Support expected envelope count in config-only delegator:
    #       Similar to TODO_54_68_18_12 (above), it should be part of _new_ `DelegatorConfigBaseEnvelopePayloadSchema`.


_config_only_func_config_dict_example = deepcopy(func_config_desc.dict_example)
_config_only_func_config_dict_example[func_envelope_][envelope_payload_].update({
    command_template_: "echo",
})

config_only_delegator_envelope_payload_desc = TypeDesc(
    dict_schema = DelegatorConfigOnlyEnvelopePayloadSchema(),
    ref_name = DelegatorConfigOnlyEnvelopePayloadSchema.__name__,
    dict_example = _config_only_func_config_dict_example,
    default_file_path = "",
)

config_only_delegator_config_desc = TypeDesc(
    dict_schema = SchemaConfigDelegatorConfigOnly(),
    ref_name = SchemaConfigDelegatorConfigOnly.__name__,
    dict_example = {
        func_configs_: {
            func_id_some_func_: config_only_delegator_envelope_payload_desc.dict_example,
        },
    },
    default_file_path = "",
)



from argrelay_api_plugin_server_abstract.DelegatorAbstract import DelegatorAbstract


class DelegatorSingleFuncAbstract(DelegatorAbstract):
    """
    This is a base delegator which implements single func.

    Unlike this (single func) class, its base `DelegatorAbstract` is potentially responsible for
    multiple funcs which is a bad pattern leading to complex logic (to handle each func differently)
    without separation of concerns.

    TODO: TODO_06_54_83_58: extend via func instead of delegator:
          Provide list of common functionality for all 1-to-1 dedicated delegators.
          For example:
          *   Each dedicated delegator should override `get_single_func_id()` which defines func it implements.
              NOTE: Some special funcs (e.g. `intercept`, `help`, ...) define `single_func_id` in their config -
              maybe this should be combined? The question is whether to exposed `single_func_id` via config or
              hide in code (as `get_single_func_id()`). Exposing via config allows matching delegator configs
              against `composite_tree` configs, but adds clutter to the config.
          *   Check common `assert func_id =` and implement it here common for all.
          *   Factor out common parts of `run_invoke_control` for simple funct
              (which only populate `custom_plugin_data`).
          *   Factor out and simplify defining func `data_envelope`.
    """



from enum import (
    auto,
    Enum,
)


class CheckEnvFunc(Enum):
    """
    Functions used by FS_36_17_84_44 `check_env`.
    """

    func_id_get_server_argrelay_version = auto()
    """
    Query server `argrelay` package version.
    """

    func_id_get_server_project_git_commit_id = auto()
    """
    Query server git commit id.
    """

    func_id_get_server_start_time = auto()
    """
    Query server start time (millis since epoch).
    """



from __future__ import annotations

from typing import Union

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory


class PluginCheckEnvOnlineMode(PluginCheckEnvAbstract):
    """
    Trivial plugin to report value of `online_mode`.
    """

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        if online_mode is None:
            result_message = "Use `offline` or `online` arg to force specific mode."
        elif online_mode:
            result_message = "Enforced `online` mode to report errors on server connection failure."
        else:
            result_message = "Enforced `offline` mode to avoid connection to server."
        return [CheckEnvResult(
            result_category = ResultCategory.VerificationSuccess,
            result_key = "online_mode",
            result_value = str(online_mode),
            result_message = result_message,
        )]



from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_client.client_command_remote.ClientCommandRemoteAbstract import ClientCommandRemoteAbstract
from argrelay_app_client.client_pipeline.BytesDstReceiver import BytesDstReceiver
from argrelay_app_client.client_pipeline.BytesHandlerAbstract import BytesHandlerAbstract
from argrelay_lib_root.enum_desc.ProcRole import ProcRole


class ClientCommandRemoteSpinner(ClientCommandRemoteAbstract):
    """
    Command for `ProcRole.ParentProcSpinner` side.

    It does not make any remote request - it only processes data out of the pipe from `ProcRole.ChildProcWorker`.
    """

    def __init__(
        self,
        call_ctx: CallContext,
        proc_role: ProcRole,
        bytes_handler: BytesHandlerAbstract,
    ):
        super().__init__(
            call_ctx,
            proc_role,
        )
        self.bytes_dst = BytesDstReceiver(
            bytes_handler,
        )

    def _execute_remote_call(
        self,
    ):
        """
        `ProcRole.ParentProcSpinner` receives data via pipe from `ProcRole.ChildProcWorker` or its exit code.
        """
        self.bytes_dst.receive_bytes()



from __future__ import annotations

import os
from typing import Union

from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.enum_desc.TermColor import TermColor
from argrelay_lib_root.misc_helper_common import (
    eprint,
    get_argrelay_dir,
)
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc


def load_client_plugin_config(
    plugin_instance_id: str,
    type_desc: TypeDesc,
) -> Union[dict, None]:
    """
    This function loads plugin config on client side.

    See FS_83_23_99_90 client plugin config override.
    """

    config_file_path = f"{get_argrelay_dir()}/conf/plugin_config/{plugin_instance_id}.yaml"
    if os.path.isfile(config_file_path):
        return type_desc.dict_from_yaml_file(config_file_path)
    else:
        return None


def filter_remaining_args(
    invocation_input: InvocationInput,
) -> list[str]:
    """
    Returns list of all unconsumed args from all FS_97_64_39_94 `token_bucket`s.

    TODO: TODO_66_09_41_16: clarify command line processing
          What this function really returns is list of unconsumed tokens.
          Consider rename.
    """

    remaining_args = []
    for i, curr_token in enumerate(invocation_input.all_tokens):
        if (
            (i not in invocation_input.consumed_token_ipos_list())
            and
            (i not in invocation_input.excluded_tokens)
        ):
            remaining_args.append(curr_token)
    return remaining_args


def prohibit_unconsumed_args(
    ii: InvocationInput,
) -> None:
    """
    TODO: TODO_86_57_50_38: Common delegator behavior: consider using delegator/func config to cause this error.
          Also, it can be caused on the server-side with redirection to DelegatorError
          (so that clients do not call that explicitly).
    """

    unconsumed_args = filter_remaining_args(ii)
    if len(unconsumed_args) > 0:
        eprint(
            f"ERROR: this function prohibits unrecognized args " +
            f"(see {TermColor.remaining_token.value}highlighted{TermColor.reset_style.value} on Alt+Shift+Q results): " +
            f"{TermColor.remaining_token.value}" +
            " ".join(unconsumed_args) +
            f"{TermColor.reset_style.value}"
        )
        exit(ClientExitCode.GeneralError.value)



from __future__ import annotations

from enum import (
    auto,
    Enum,
)

from argrelay_lib_root.enum_desc.CompType import CompType


class CallConv(Enum):
    """
    `CallConv` defines calling convention between Shell and `argrelay` client.

    `CallConv` is closely related to `RunMode`.

    It tells how data is passed from Shell to argrelay client (similar concept to C/C++ calling conventions):
    *   `CallConv.CliArgsConv` is used for `RunMode.Invocation` to pass data via command line args.
    *   `CallConv.EnvVarsConv` is used for `RunMode.Completion` to pass data via env vars.
    """

    CliArgsConv = auto()
    EnvVarsConv = auto()

    def __str__(self):
        return self.name

    @classmethod
    def from_comp_type(
        cls,
        comp_type: CompType,
    ) -> CallConv:
        """
        Derive `CallConv` on client side from `CompType`.
        """

        if comp_type is CompType.InvokeAction:
            return CallConv.CliArgsConv
        else:
            return CallConv.EnvVarsConv



from argrelay_api_server_cli.server_spec.CallContext import CallContext


class ClientCommandAbstract:

    def __init__(
        self,
        call_ctx: CallContext,
    ):
        self.call_ctx: CallContext = call_ctx

    def execute_command(
        self,
    ):
        pass



from __future__ import annotations

import json
import os
import subprocess

from argrelay_api_plugin_server_abstract.delegator_utils import (
    clean_prop_value,
)
from argrelay_api_plugin_server_abstract.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_lib_server_plugin_core.plugin_delegator.client_invocation_utils import prohibit_unconsumed_args
from argrelay_lib_server_plugin_demo.demo_service.ServicePropName import ServicePropName
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import populate_search_control

func_id_ssh_dst_ = "func_id_ssh_dst"

class_ssh_dst_ = "class_ssh_dst"

ssh_dst_container_ipos_ = 1


class DelegatorSshDst(DelegatorSingleFuncAbstract):
    """
    This is a demo delegator wrapping `ssh` command.

    Unlike naked `ssh`, wrapped `ssh` allows selecting destinations by
    user-specific metadata properties - see `search_control_list_`.

    It reuses metadata properties defined in `ServicePropName` simply to
    avoid re-defining something similar for demo purposes, but it is a standalone
    plugin otherwise.

    It is possible to categorize `ssh` destination by any of `ServicePropName`-s.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: func_id_ssh_dst_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        populate_search_control(
                            collection_name = class_ssh_dst_,
                            props_to_values_dict = {
                                ReservedPropName.envelope_class.name: class_ssh_dst_,
                            },
                            arg_name_to_prop_name_map = [
                                # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                                {"class": ReservedPropName.envelope_class.name},
                                # ---
                                {"code": ServicePropName.code_maturity.name},
                                {"stage": ServicePropName.flow_stage.name},
                                {"region": ServicePropName.geo_region.name},
                                {"cluster": ServicePropName.cluster_name.name},
                                # ---
                                {"group": ServicePropName.group_label.name},
                                {"service": ServicePropName.service_name.name},
                                {"mode": ServicePropName.run_mode.name},
                                # ---
                                {"user": ServicePropName.user_name.name},
                                {"host": ServicePropName.host_name.name},
                                {"path": ServicePropName.dir_path.name},
                                # ---
                                {"status": ServicePropName.live_status.name},
                                {"dc": ServicePropName.data_center.name},
                                {"ip": ServicePropName.ip_address.name},
                            ],
                        ),
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
                ReservedPropName.help_hint.name: "Use `ssh` to log into the destination.",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_ssh_dst_,
            },
        ]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_ssh_dst_

        # Even if these functions do not support varargs, query all - let client decide what to do.
        # Search `data_envelope`-s based on existing args on command line:
        vararg_container = interp_ctx.envelope_containers[ssh_dst_container_ipos_]
        vararg_container.data_envelopes = (
            local_server
            .get_query_engine()
            .query_data_envelopes_for(vararg_container)
        )

        # Plugin to invoke on client side:
        delegator_plugin_instance_id = self.plugin_instance_id
        # Package into `InvocationInput` payload object:
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == func_id_ssh_dst_

        prohibit_unconsumed_args(invocation_input)

        ssh_dst_data_envelopes = invocation_input.envelope_containers[ssh_dst_container_ipos_].data_envelopes

        if len(ssh_dst_data_envelopes) > 1:
            for ssh_dst_data_envelope in ssh_dst_data_envelopes:
                eprint(_data_envelope_to_str(ssh_dst_data_envelope))

            # TODO: TODO_20_61_16_31 `cardinality_hook`: run different funcs based on `data_envelope` set size
            eprint(
                "ERROR: `ssh` destination is ambiguous "
                "(multiple candidates based on given command line input)"
            )
            exit(ClientExitCode.GeneralError.value)

        elif len(ssh_dst_data_envelopes) == 0:
            eprint("ERROR: `ssh` destination not found based on given command line input.")
            exit(ClientExitCode.GeneralError.value)

        else:
            ssh_dst_data_envelope = ssh_dst_data_envelopes[0]
            _run_ssh(ssh_dst_data_envelope)


def _data_envelope_to_str(
    data_envelope: dict,
) -> str:
    return json.dumps(
        data_envelope,
        indent = 4,
    )


def _run_ssh(
    ssh_dst_data_envelope: dict,
) -> None:
    eprint(_data_envelope_to_str(ssh_dst_data_envelope))

    user_name = clean_prop_value(ssh_dst_data_envelope[ServicePropName.user_name.name])
    if not user_name:
        user_name = os.getlogin()

    host_name = clean_prop_value(ssh_dst_data_envelope[ServicePropName.host_name.name])
    dir_path = clean_prop_value(ssh_dst_data_envelope[ServicePropName.dir_path.name])

    ssh_dst = f"{user_name}@{host_name}:{dir_path}"
    eprint(f"INFO: starting shell via `ssh` at destination: {ssh_dst}")

    # Run `ssh` by starting shell explicitly to select required target dir:
    sub_proc = subprocess.run(
        [
            "ssh",
            "-t",
            f"{user_name}@{host_name}",
            f"cd \"{dir_path}\" ; bash --login",
        ],
    )
    exit_code = sub_proc.returncode
    exit(exit_code)



from __future__ import annotations

import json

from argrelay_api_plugin_server_abstract.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_core.plugin_delegator.client_invocation_utils import prohibit_unconsumed_args
from argrelay_lib_server_plugin_demo.demo_service.DelegatorServiceHostBase import (
    DelegatorServiceHostBase,
    get_host_search_control,
)
from argrelay_lib_server_plugin_demo.demo_service.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)

func_id_list_host_ = "func_id_list_host"

host_container_ipos_ = 1


class DelegatorServiceHostList(DelegatorServiceHostBase):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        host_search_control = get_host_search_control()

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: func_id_list_host_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
                ReservedPropName.help_hint.name: "List remote hosts matching search query",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_list_host_,
            },
        ]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_list_host_

        vararg_container_ipos = host_container_ipos_

        return self._compose_invocation_input_for_list(
            interp_ctx,
            local_server,
            vararg_container_ipos,
            ServiceEnvelopeClass.class_host.name,
        )

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        """
        Print `data_envelope`-s received from server on client side.
        """

        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == func_id_list_host_

        prohibit_unconsumed_args(invocation_input)

        for data_envelope in invocation_input.envelope_containers[host_container_ipos_].data_envelopes:
            print(json.dumps(data_envelope))



from __future__ import annotations

from marshmallow import (
    fields,
    RAISE,
)

from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_root.schema_config.ConnectionConfigSchema import connection_config_desc
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig
from argrelay_schema_config_server.schema_config_server_app.GuiBannerConfigSchema import gui_banner_config_desc
from argrelay_schema_config_server.schema_config_server_app.MongoConfigSchema import mongo_config_desc
from argrelay_schema_config_server.schema_config_server_app.QueryCacheConfigSchema import query_cache_config_desc
from argrelay_schema_config_server.schema_config_server_app.ServerPluginControlSchema import server_plugin_control_desc

connection_config_ = "connection_config"
mongo_config_ = "mongo_config"
query_cache_config_ = "query_cache_config"
gui_banner_config_ = "gui_banner_config"
default_gui_command_ = "default_gui_command"
server_plugin_control_ = "server_plugin_control"


class ServerConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = ServerConfig

    connection_config = fields.Nested(
        connection_config_desc.dict_schema,
        required = True,
    )

    mongo_config = fields.Nested(
        mongo_config_desc.dict_schema,
        required = True,
    )

    query_cache_config = fields.Nested(
        query_cache_config_desc.dict_schema,
        required = True,
    )

    gui_banner_config = fields.Nested(
        gui_banner_config_desc.dict_schema,
        required = True,
    )

    default_gui_command = fields.String(
        required = False,
        load_default = None,
    )

    server_plugin_control = fields.Nested(
        server_plugin_control_desc.dict_schema,
        required = True,
    )


server_config_desc = TypeDesc(
    dict_schema = ServerConfigSchema(),
    ref_name = ServerConfigSchema.__name__,
    dict_example = {
        connection_config_: connection_config_desc.dict_example,
        mongo_config_: mongo_config_desc.dict_example,
        query_cache_config_: query_cache_config_desc.dict_example,
        gui_banner_config_: gui_banner_config_desc.dict_example,
        server_plugin_control_: server_plugin_control_desc.dict_example,
    },
    default_file_path = "argrelay_server.yaml",
)



from enum import (
    auto,
    Enum,
)


class NoDataPropName(Enum):
    """
    These `prop_name`-s are used for (see also):
    *   `ReservedEnvelopeClass.class_no_data`
    *   `SpecialFunc.func_id_no_data`
    """

    no_data_prop_name_one = auto()
    no_data_prop_name_two = auto()



from dataclasses import (
    dataclass,
    field,
)

from argrelay_lib_root.enum_desc.ValueSource import ValueSource


@dataclass
class AssignedValue:
    prop_value: str = field()
    value_source: ValueSource = field()



from __future__ import annotations

from argrelay_api_plugin_server_abstract.AbstractLoader import AbstractLoader
from argrelay_app_server.relay_server.IndexModel import IndexModel
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_core.enum_desc.NoDataPropName import NoDataPropName


class NoopLoader(AbstractLoader):

    def list_index_models(
        self,
    ) -> list[IndexModel]:
        return [
            IndexModel(
                collection_name = ReservedEnvelopeClass.class_no_data.name,
                index_props = (
                    # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                    [ReservedPropName.envelope_class.name] +
                    [e.name for e in NoDataPropName]
                )
            ),
        ]
